"""
Enterprise Plugin System - Extensible plugin architecture.

Enables modular extension of the framework through plugins
that can add new agents, tools, workflows, and more.

Features:
- Plugin discovery and loading
- Plugin lifecycle management
- Plugin dependencies
- Hot reload support
- Plugin marketplace
"""

import asyncio
import importlib
import importlib.util
import inspect
import logging
import os
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


# =============================================================================
# Plugin Status
# =============================================================================

class PluginStatus(Enum):
    """Plugin status states."""
    DISCOVERED = "discovered"
    LOADED = "loaded"
    INITIALIZED = "initialized"
    ACTIVE = "active"
    DISABLED = "disabled"
    FAILED = "failed"
    UNLOADED = "unloaded"


# =============================================================================
# Plugin Metadata
# =============================================================================

@dataclass
class PluginMetadata:
    """Plugin metadata and configuration."""
    name: str
    version: str
    description: str = ""
    author: str = ""
    email: str = ""
    url: str = ""
    license: str = ""
    
    # Dependencies
    dependencies: List[str] = field(default_factory=list)
    python_requires: str = ">=3.9"
    
    # Configuration
    config_schema: Optional[Dict] = None
    default_config: Dict = field(default_factory=dict)
    
    # Capabilities
    provides_agents: bool = False
    provides_tools: bool = False
    provides_workflows: bool = False
    provides_adapters: bool = False
    provides_middleware: bool = False
    
    # Tags for discovery
    tags: List[str] = field(default_factory=list)


# =============================================================================
# Plugin Base Class
# =============================================================================

class Plugin(ABC):
    """
    Base class for all plugins.
    
    Usage:
        >>> class MyPlugin(Plugin):
        ...     metadata = PluginMetadata(
        ...         name="my-plugin",
        ...         version="1.0.0",
        ...         description="A sample plugin",
        ...     )
        ...     
        ...     async def initialize(self, config: Dict):
        ...         self.config = config
        ...     
        ...     async def activate(self):
        ...         # Register agents, tools, etc.
        ...         pass
        ...     
        ...     async def deactivate(self):
        ...         # Cleanup
        ...         pass
    """
    
    metadata: PluginMetadata
    status: PluginStatus = PluginStatus.DISCOVERED
    config: Dict = {}
    
    @abstractmethod
    async def initialize(self, config: Dict[str, Any]):
        """Initialize the plugin with configuration."""
        pass
    
    @abstractmethod
    async def activate(self):
        """Activate the plugin and register components."""
        pass
    
    @abstractmethod
    async def deactivate(self):
        """Deactivate and cleanup the plugin."""
        pass
    
    def get_agents(self) -> List[Type]:
        """Get agents provided by this plugin."""
        return []
    
    def get_tools(self) -> List[Type]:
        """Get tools provided by this plugin."""
        return []
    
    def get_workflows(self) -> List[Type]:
        """Get workflows provided by this plugin."""
        return []
    
    def get_middleware(self) -> List[Type]:
        """Get middleware provided by this plugin."""
        return []


# =============================================================================
# Plugin Registry
# =============================================================================

@dataclass
class PluginInfo:
    """Information about a registered plugin."""
    plugin: Plugin
    module: Any
    path: Optional[Path]
    status: PluginStatus = PluginStatus.DISCOVERED
    loaded_at: Optional[datetime] = None
    error: Optional[str] = None


class PluginRegistry:
    """
    Registry for managing plugins.
    
    Usage:
        >>> registry = PluginRegistry()
        >>> registry.discover("./plugins")
        >>> await registry.load_all()
        >>> await registry.activate_all()
    """
    
    def __init__(self):
        self._plugins: Dict[str, PluginInfo] = {}
        self._hooks: Dict[str, List[Callable]] = {}
        self._lock = asyncio.Lock()
    
    def register(self, plugin: Plugin, module: Any = None, path: Path = None):
        """Register a plugin."""
        name = plugin.metadata.name
        
        self._plugins[name] = PluginInfo(
            plugin=plugin,
            module=module,
            path=path,
            status=PluginStatus.DISCOVERED,
        )
        
        logger.info(f"Registered plugin: {name} v{plugin.metadata.version}")
    
    def unregister(self, name: str):
        """Unregister a plugin."""
        if name in self._plugins:
            del self._plugins[name]
            logger.info(f"Unregistered plugin: {name}")
    
    def get(self, name: str) -> Optional[Plugin]:
        """Get a plugin by name."""
        info = self._plugins.get(name)
        return info.plugin if info else None
    
    def get_all(self) -> List[Plugin]:
        """Get all registered plugins."""
        return [info.plugin for info in self._plugins.values()]
    
    def get_by_tag(self, tag: str) -> List[Plugin]:
        """Get plugins by tag."""
        return [
            info.plugin
            for info in self._plugins.values()
            if tag in info.plugin.metadata.tags
        ]
    
    def get_active(self) -> List[Plugin]:
        """Get all active plugins."""
        return [
            info.plugin
            for info in self._plugins.values()
            if info.status == PluginStatus.ACTIVE
        ]
    
    def discover(self, path: str):
        """
        Discover plugins in a directory.
        
        Looks for:
        - Python packages with plugin.py
        - Python files with Plugin subclass
        - plugin.json/plugin.yaml configs
        """
        plugin_path = Path(path)
        
        if not plugin_path.exists():
            logger.warning(f"Plugin path does not exist: {path}")
            return
        
        # Add to Python path
        if str(plugin_path) not in sys.path:
            sys.path.insert(0, str(plugin_path))
        
        for item in plugin_path.iterdir():
            try:
                if item.is_dir():
                    self._discover_package(item)
                elif item.suffix == ".py":
                    self._discover_file(item)
            except Exception as e:
                logger.error(f"Error discovering plugin {item}: {e}")
    
    def _discover_package(self, path: Path):
        """Discover plugin from package directory."""
        plugin_file = path / "plugin.py"
        init_file = path / "__init__.py"
        
        target = plugin_file if plugin_file.exists() else init_file
        
        if target.exists():
            self._load_plugin_from_file(target, path)
    
    def _discover_file(self, path: Path):
        """Discover plugin from single file."""
        if path.name.startswith("_"):
            return
        
        self._load_plugin_from_file(path, path.parent)
    
    def _load_plugin_from_file(self, path: Path, plugin_path: Path):
        """Load a plugin from a Python file."""
        spec = importlib.util.spec_from_file_location(
            path.stem,
            path,
        )
        
        if not spec or not spec.loader:
            return
        
        module = importlib.util.module_from_spec(spec)
        
        try:
            spec.loader.exec_module(module)
        except Exception as e:
            logger.error(f"Error loading module {path}: {e}")
            return
        
        # Find Plugin subclasses
        for name, obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(obj, Plugin) and obj is not Plugin:
                try:
                    plugin = obj()
                    self.register(plugin, module, plugin_path)
                except Exception as e:
                    logger.error(f"Error instantiating plugin {name}: {e}")
    
    async def load(self, name: str, config: Dict = None):
        """Load and initialize a specific plugin."""
        info = self._plugins.get(name)
        
        if not info:
            raise PluginError(f"Plugin not found: {name}")
        
        if info.status in (PluginStatus.LOADED, PluginStatus.ACTIVE):
            return
        
        # Check dependencies
        for dep in info.plugin.metadata.dependencies:
            if dep not in self._plugins:
                raise PluginError(f"Missing dependency: {dep}")
            
            dep_info = self._plugins[dep]
            if dep_info.status not in (PluginStatus.LOADED, PluginStatus.ACTIVE):
                await self.load(dep, config)
        
        try:
            # Merge config with defaults
            plugin_config = {
                **info.plugin.metadata.default_config,
                **(config or {}),
            }
            
            await info.plugin.initialize(plugin_config)
            info.status = PluginStatus.LOADED
            info.loaded_at = datetime.now()
            
            logger.info(f"Loaded plugin: {name}")
            
        except Exception as e:
            info.status = PluginStatus.FAILED
            info.error = str(e)
            raise PluginError(f"Failed to load plugin {name}: {e}")
    
    async def load_all(self, config: Dict = None):
        """Load all discovered plugins."""
        for name in list(self._plugins.keys()):
            try:
                await self.load(name, config)
            except PluginError as e:
                logger.error(f"Error loading plugin: {e}")
    
    async def activate(self, name: str):
        """Activate a loaded plugin."""
        info = self._plugins.get(name)
        
        if not info:
            raise PluginError(f"Plugin not found: {name}")
        
        if info.status == PluginStatus.ACTIVE:
            return
        
        if info.status != PluginStatus.LOADED:
            raise PluginError(f"Plugin not loaded: {name}")
        
        try:
            await info.plugin.activate()
            info.status = PluginStatus.ACTIVE
            
            # Trigger hooks
            await self._trigger_hook("plugin.activated", info.plugin)
            
            logger.info(f"Activated plugin: {name}")
            
        except Exception as e:
            info.status = PluginStatus.FAILED
            info.error = str(e)
            raise PluginError(f"Failed to activate plugin {name}: {e}")
    
    async def activate_all(self):
        """Activate all loaded plugins."""
        for name, info in self._plugins.items():
            if info.status == PluginStatus.LOADED:
                try:
                    await self.activate(name)
                except PluginError as e:
                    logger.error(f"Error activating plugin: {e}")
    
    async def deactivate(self, name: str):
        """Deactivate an active plugin."""
        info = self._plugins.get(name)
        
        if not info:
            raise PluginError(f"Plugin not found: {name}")
        
        if info.status != PluginStatus.ACTIVE:
            return
        
        try:
            await info.plugin.deactivate()
            info.status = PluginStatus.LOADED
            
            await self._trigger_hook("plugin.deactivated", info.plugin)
            
            logger.info(f"Deactivated plugin: {name}")
            
        except Exception as e:
            info.error = str(e)
            raise PluginError(f"Failed to deactivate plugin {name}: {e}")
    
    async def deactivate_all(self):
        """Deactivate all active plugins."""
        for name, info in self._plugins.items():
            if info.status == PluginStatus.ACTIVE:
                try:
                    await self.deactivate(name)
                except PluginError as e:
                    logger.error(f"Error deactivating plugin: {e}")
    
    async def reload(self, name: str):
        """Hot reload a plugin."""
        info = self._plugins.get(name)
        
        if not info:
            raise PluginError(f"Plugin not found: {name}")
        
        was_active = info.status == PluginStatus.ACTIVE
        config = info.plugin.config
        
        # Deactivate if active
        if was_active:
            await self.deactivate(name)
        
        # Reload module
        if info.module:
            importlib.reload(info.module)
            
            # Re-find plugin class
            for obj_name, obj in inspect.getmembers(info.module, inspect.isclass):
                if issubclass(obj, Plugin) and obj is not Plugin:
                    info.plugin = obj()
                    break
        
        # Re-load and activate
        await self.load(name, config)
        
        if was_active:
            await self.activate(name)
        
        logger.info(f"Reloaded plugin: {name}")
    
    def on_hook(self, hook_name: str):
        """Decorator to register a plugin hook."""
        def decorator(fn: Callable):
            if hook_name not in self._hooks:
                self._hooks[hook_name] = []
            self._hooks[hook_name].append(fn)
            return fn
        return decorator
    
    async def _trigger_hook(self, hook_name: str, *args, **kwargs):
        """Trigger all handlers for a hook."""
        handlers = self._hooks.get(hook_name, [])
        
        for handler in handlers:
            try:
                result = handler(*args, **kwargs)
                if asyncio.iscoroutine(result):
                    await result
            except Exception as e:
                logger.error(f"Hook handler error: {e}")


# =============================================================================
# Plugin Loader
# =============================================================================

class PluginLoader:
    """
    Loads plugins from various sources.
    
    Supports:
    - Local directories
    - Git repositories
    - PyPI packages
    - URLs
    """
    
    def __init__(self, registry: PluginRegistry):
        self.registry = registry
        self.cache_dir = Path.home() / ".agenticai" / "plugins"
    
    async def load_from_path(self, path: str):
        """Load plugins from a local path."""
        self.registry.discover(path)
    
    async def load_from_git(self, url: str, ref: str = "main"):
        """Load plugins from a git repository."""
        import subprocess
        
        # Create cache directory
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        
        # Clone or update repo
        repo_name = url.split("/")[-1].replace(".git", "")
        repo_path = self.cache_dir / repo_name
        
        if repo_path.exists():
            subprocess.run(
                ["git", "-C", str(repo_path), "pull"],
                check=True,
            )
        else:
            subprocess.run(
                ["git", "clone", url, str(repo_path)],
                check=True,
            )
        
        subprocess.run(
            ["git", "-C", str(repo_path), "checkout", ref],
            check=True,
        )
        
        self.registry.discover(str(repo_path))
    
    async def load_from_pip(self, package: str):
        """Load plugins from a pip package."""
        import subprocess
        
        # Install package
        subprocess.run(
            [sys.executable, "-m", "pip", "install", package],
            check=True,
        )
        
        # Try to import and discover plugins
        try:
            module = importlib.import_module(package)
            module_path = Path(module.__file__).parent
            self.registry.discover(str(module_path))
        except ImportError as e:
            raise PluginError(f"Failed to import package: {e}")


# =============================================================================
# Plugin Manager
# =============================================================================

class PluginManager:
    """
    High-level plugin management.
    
    Usage:
        >>> manager = PluginManager()
        >>> 
        >>> # Load plugins
        >>> await manager.load_from("./plugins")
        >>> await manager.load_from("https://github.com/user/plugin.git")
        >>> 
        >>> # Activate all
        >>> await manager.activate_all()
        >>> 
        >>> # Get components
        >>> agents = manager.get_all_agents()
        >>> tools = manager.get_all_tools()
    """
    
    def __init__(self, config: Dict = None):
        self.registry = PluginRegistry()
        self.loader = PluginLoader(self.registry)
        self.config = config or {}
    
    async def load_from(self, source: str, **kwargs):
        """Load plugins from any source."""
        if source.startswith("http") or source.endswith(".git"):
            await self.loader.load_from_git(source, **kwargs)
        elif "==" in source or source.startswith("pip:"):
            package = source.replace("pip:", "")
            await self.loader.load_from_pip(package)
        else:
            await self.loader.load_from_path(source)
        
        # Load discovered plugins
        await self.registry.load_all(self.config)
    
    async def activate_all(self):
        """Activate all loaded plugins."""
        await self.registry.activate_all()
    
    async def deactivate_all(self):
        """Deactivate all plugins."""
        await self.registry.deactivate_all()
    
    def get_all_agents(self) -> List[Type]:
        """Get all agents from active plugins."""
        agents = []
        for plugin in self.registry.get_active():
            agents.extend(plugin.get_agents())
        return agents
    
    def get_all_tools(self) -> List[Type]:
        """Get all tools from active plugins."""
        tools = []
        for plugin in self.registry.get_active():
            tools.extend(plugin.get_tools())
        return tools
    
    def get_all_workflows(self) -> List[Type]:
        """Get all workflows from active plugins."""
        workflows = []
        for plugin in self.registry.get_active():
            workflows.extend(plugin.get_workflows())
        return workflows
    
    def get_all_middleware(self) -> List[Type]:
        """Get all middleware from active plugins."""
        middleware = []
        for plugin in self.registry.get_active():
            middleware.extend(plugin.get_middleware())
        return middleware


# =============================================================================
# Decorators
# =============================================================================

def plugin(
    name: str,
    version: str,
    description: str = "",
    **kwargs,
):
    """
    Decorator to create a simple plugin.
    
    Usage:
        >>> @plugin("my-plugin", "1.0.0", description="A plugin")
        ... class MyPlugin:
        ...     def activate(self):
        ...         pass
    """
    def decorator(cls):
        # Create metadata
        metadata = PluginMetadata(
            name=name,
            version=version,
            description=description,
            **kwargs,
        )
        
        # Create plugin class
        class WrappedPlugin(Plugin):
            metadata = metadata
            
            def __init__(self):
                self._impl = cls()
            
            async def initialize(self, config):
                self.config = config
                if hasattr(self._impl, "initialize"):
                    result = self._impl.initialize(config)
                    if asyncio.iscoroutine(result):
                        await result
            
            async def activate(self):
                if hasattr(self._impl, "activate"):
                    result = self._impl.activate()
                    if asyncio.iscoroutine(result):
                        await result
            
            async def deactivate(self):
                if hasattr(self._impl, "deactivate"):
                    result = self._impl.deactivate()
                    if asyncio.iscoroutine(result):
                        await result
            
            def get_agents(self):
                if hasattr(self._impl, "get_agents"):
                    return self._impl.get_agents()
                return []
            
            def get_tools(self):
                if hasattr(self._impl, "get_tools"):
                    return self._impl.get_tools()
                return []
        
        # Auto-register
        global_plugin_registry.register(WrappedPlugin())
        
        return cls
    return decorator


# =============================================================================
# Global Registry
# =============================================================================

global_plugin_registry = PluginRegistry()


def get_plugin_registry() -> PluginRegistry:
    """Get the global plugin registry."""
    return global_plugin_registry


def register_plugin(plugin: Plugin):
    """Register a plugin in the global registry."""
    global_plugin_registry.register(plugin)


def get_plugin(name: str) -> Optional[Plugin]:
    """Get a plugin by name."""
    return global_plugin_registry.get(name)


# =============================================================================
# Errors
# =============================================================================

class PluginError(Exception):
    """Plugin-related error."""
    pass



"""
Enterprise Value Object Module.

Provides immutable value objects, equality by value,
and self-validation for DDD architectures.

Example:
    # Define a value object
    @value_object
    class Money:
        amount: Decimal
        currency: str
        
        def add(self, other: "Money") -> "Money":
            if self.currency != other.currency:
                raise ValueError("Currency mismatch")
            return Money(self.amount + other.amount, self.currency)
    
    # Use value object
    price = Money(Decimal("99.99"), "USD")
    tax = Money(Decimal("8.00"), "USD")
    total = price.add(tax)
"""

from __future__ import annotations

import copy
import hashlib
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields as dataclass_fields
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from functools import total_ordering
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

T = TypeVar('T')
V = TypeVar('V', bound='ValueObject')


class ValueObjectError(Exception):
    """Value object error."""
    pass


class ValidationError(ValueObjectError):
    """Validation error."""
    pass


class ImmutabilityError(ValueObjectError):
    """Immutability violation error."""
    pass


class ValueObject(ABC):
    """
    Base class for value objects.
    
    Value objects are defined by their attributes and are immutable.
    Two value objects are equal if all their attributes are equal.
    """
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._get_equality_components() == other._get_equality_components()
    
    def __hash__(self) -> int:
        return hash(self._get_equality_components())
    
    @abstractmethod
    def _get_equality_components(self) -> tuple:
        """Get components for equality comparison."""
        pass
    
    def copy_with(self, **changes: Any) -> "ValueObject":
        """Create a copy with changes."""
        # Must be implemented by subclasses
        raise NotImplementedError


@dataclass(frozen=True)
class SimpleValueObject(ValueObject):
    """
    Simple value object using dataclass.
    
    Example:
        @dataclass(frozen=True)
        class Email(SimpleValueObject):
            address: str
            
            def __post_init__(self):
                if "@" not in self.address:
                    raise ValidationError("Invalid email")
    """
    
    def _get_equality_components(self) -> tuple:
        return tuple(
            getattr(self, f.name)
            for f in dataclass_fields(self)
        )
    
    def copy_with(self, **changes: Any) -> "SimpleValueObject":
        """Create a copy with changes."""
        current = {
            f.name: getattr(self, f.name)
            for f in dataclass_fields(self)
        }
        current.update(changes)
        return self.__class__(**current)


# Common Value Objects
@dataclass(frozen=True)
@total_ordering
class Money(SimpleValueObject):
    """Money value object."""
    amount: Decimal
    currency: str = "USD"
    
    def __post_init__(self):
        if not isinstance(self.amount, Decimal):
            object.__setattr__(self, 'amount', Decimal(str(self.amount)))
    
    def __lt__(self, other: "Money") -> bool:
        if not isinstance(other, Money):
            return NotImplemented
        if self.currency != other.currency:
            raise ValueObjectError("Cannot compare different currencies")
        return self.amount < other.amount
    
    def add(self, other: "Money") -> "Money":
        """Add money amounts."""
        if self.currency != other.currency:
            raise ValueObjectError("Currency mismatch")
        return Money(self.amount + other.amount, self.currency)
    
    def subtract(self, other: "Money") -> "Money":
        """Subtract money amounts."""
        if self.currency != other.currency:
            raise ValueObjectError("Currency mismatch")
        return Money(self.amount - other.amount, self.currency)
    
    def multiply(self, factor: Union[int, float, Decimal]) -> "Money":
        """Multiply by a factor."""
        return Money(
            self.amount * Decimal(str(factor)),
            self.currency,
        )
    
    def is_positive(self) -> bool:
        """Check if amount is positive."""
        return self.amount > 0
    
    def is_zero(self) -> bool:
        """Check if amount is zero."""
        return self.amount == 0
    
    @classmethod
    def zero(cls, currency: str = "USD") -> "Money":
        """Create zero money."""
        return cls(Decimal("0"), currency)
    
    def __str__(self) -> str:
        return f"{self.currency} {self.amount:.2f}"


@dataclass(frozen=True)
class EmailAddress(SimpleValueObject):
    """Email address value object."""
    address: str
    
    def __post_init__(self):
        if not self.address or "@" not in self.address:
            raise ValidationError(f"Invalid email address: {self.address}")
        
        local, domain = self.address.rsplit("@", 1)
        if not local or not domain or "." not in domain:
            raise ValidationError(f"Invalid email address: {self.address}")
    
    @property
    def local_part(self) -> str:
        return self.address.rsplit("@", 1)[0]
    
    @property
    def domain(self) -> str:
        return self.address.rsplit("@", 1)[1]
    
    def __str__(self) -> str:
        return self.address


@dataclass(frozen=True)
class PhoneNumber(SimpleValueObject):
    """Phone number value object."""
    number: str
    country_code: str = "+1"
    
    def __post_init__(self):
        # Clean the number
        cleaned = "".join(c for c in self.number if c.isdigit())
        if len(cleaned) < 7:
            raise ValidationError(f"Invalid phone number: {self.number}")
        object.__setattr__(self, 'number', cleaned)
    
    @property
    def formatted(self) -> str:
        if len(self.number) == 10:
            return f"{self.country_code} ({self.number[:3]}) {self.number[3:6]}-{self.number[6:]}"
        return f"{self.country_code} {self.number}"
    
    def __str__(self) -> str:
        return self.formatted


@dataclass(frozen=True)
class Address(SimpleValueObject):
    """Address value object."""
    street: str
    city: str
    state: str
    postal_code: str
    country: str = "USA"
    apartment: Optional[str] = None
    
    def __post_init__(self):
        if not self.street or not self.city:
            raise ValidationError("Street and city are required")
    
    @property
    def full_address(self) -> str:
        parts = [self.street]
        if self.apartment:
            parts.append(f"Apt {self.apartment}")
        parts.append(f"{self.city}, {self.state} {self.postal_code}")
        parts.append(self.country)
        return "\n".join(parts)
    
    def __str__(self) -> str:
        return f"{self.street}, {self.city}, {self.state} {self.postal_code}"


@dataclass(frozen=True)
class DateRange(SimpleValueObject):
    """Date range value object."""
    start: date
    end: date
    
    def __post_init__(self):
        if self.end < self.start:
            raise ValidationError("End date must be after start date")
    
    @property
    def days(self) -> int:
        return (self.end - self.start).days
    
    def contains(self, d: date) -> bool:
        """Check if date is in range."""
        return self.start <= d <= self.end
    
    def overlaps(self, other: "DateRange") -> bool:
        """Check if ranges overlap."""
        return self.start <= other.end and other.start <= self.end
    
    def merge(self, other: "DateRange") -> "DateRange":
        """Merge overlapping ranges."""
        if not self.overlaps(other):
            raise ValueObjectError("Ranges do not overlap")
        return DateRange(
            min(self.start, other.start),
            max(self.end, other.end),
        )


@dataclass(frozen=True)
class Percentage(SimpleValueObject):
    """Percentage value object."""
    value: Decimal
    
    def __post_init__(self):
        if not isinstance(self.value, Decimal):
            object.__setattr__(self, 'value', Decimal(str(self.value)))
        if self.value < 0 or self.value > 100:
            raise ValidationError("Percentage must be between 0 and 100")
    
    def apply_to(self, amount: Decimal) -> Decimal:
        """Apply percentage to an amount."""
        return amount * (self.value / 100)
    
    @classmethod
    def from_fraction(cls, fraction: float) -> "Percentage":
        """Create from fraction (0-1)."""
        return cls(Decimal(str(fraction * 100)))
    
    def __str__(self) -> str:
        return f"{self.value}%"


@dataclass(frozen=True)
class Quantity(SimpleValueObject):
    """Quantity value object."""
    value: int
    unit: str = "pcs"
    
    def __post_init__(self):
        if self.value < 0:
            raise ValidationError("Quantity cannot be negative")
    
    def add(self, other: "Quantity") -> "Quantity":
        """Add quantities."""
        if self.unit != other.unit:
            raise ValueObjectError("Unit mismatch")
        return Quantity(self.value + other.value, self.unit)
    
    def subtract(self, other: "Quantity") -> "Quantity":
        """Subtract quantities."""
        if self.unit != other.unit:
            raise ValueObjectError("Unit mismatch")
        result = self.value - other.value
        if result < 0:
            raise ValueObjectError("Result would be negative")
        return Quantity(result, self.unit)
    
    def is_zero(self) -> bool:
        return self.value == 0
    
    @classmethod
    def zero(cls, unit: str = "pcs") -> "Quantity":
        return cls(0, unit)
    
    def __str__(self) -> str:
        return f"{self.value} {self.unit}"


@dataclass(frozen=True)
class Coordinate(SimpleValueObject):
    """Geographic coordinate value object."""
    latitude: float
    longitude: float
    
    def __post_init__(self):
        if not (-90 <= self.latitude <= 90):
            raise ValidationError("Latitude must be between -90 and 90")
        if not (-180 <= self.longitude <= 180):
            raise ValidationError("Longitude must be between -180 and 180")
    
    def distance_to(self, other: "Coordinate") -> float:
        """Calculate approximate distance in kilometers."""
        from math import radians, sin, cos, sqrt, atan2
        
        R = 6371  # Earth's radius in km
        
        lat1, lon1 = radians(self.latitude), radians(self.longitude)
        lat2, lon2 = radians(other.latitude), radians(other.longitude)
        
        dlat = lat2 - lat1
        dlon = lon2 - lon1
        
        a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
        c = 2 * atan2(sqrt(a), sqrt(1-a))
        
        return R * c
    
    def __str__(self) -> str:
        return f"({self.latitude}, {self.longitude})"


@dataclass(frozen=True)
class URL(SimpleValueObject):
    """URL value object."""
    value: str
    
    def __post_init__(self):
        if not self.value:
            raise ValidationError("URL cannot be empty")
        if not (self.value.startswith("http://") or self.value.startswith("https://")):
            raise ValidationError("URL must start with http:// or https://")
    
    @property
    def is_secure(self) -> bool:
        return self.value.startswith("https://")
    
    @property
    def domain(self) -> str:
        from urllib.parse import urlparse
        return urlparse(self.value).netloc
    
    def __str__(self) -> str:
        return self.value


@dataclass(frozen=True)
class Password(SimpleValueObject):
    """Password value object with hashing."""
    _hash: str
    
    @classmethod
    def create(cls, plain_text: str, min_length: int = 8) -> "Password":
        """Create password from plain text."""
        if len(plain_text) < min_length:
            raise ValidationError(f"Password must be at least {min_length} characters")
        
        # Simple hash - in production use proper password hashing
        hash_value = hashlib.sha256(plain_text.encode()).hexdigest()
        return cls(hash_value)
    
    def verify(self, plain_text: str) -> bool:
        """Verify password."""
        hash_value = hashlib.sha256(plain_text.encode()).hexdigest()
        return self._hash == hash_value
    
    def __str__(self) -> str:
        return "********"
    
    def __repr__(self) -> str:
        return "Password(***)"


class ValueCollection(ValueObject, Generic[T]):
    """
    Immutable collection value object.
    
    Example:
        tags = ValueCollection[str](["python", "ddd", "enterprise"])
    """
    
    def __init__(self, items: List[T]):
        self._items: Tuple[T, ...] = tuple(items)
    
    @property
    def items(self) -> Tuple[T, ...]:
        return self._items
    
    def __iter__(self):
        return iter(self._items)
    
    def __len__(self) -> int:
        return len(self._items)
    
    def __contains__(self, item: T) -> bool:
        return item in self._items
    
    def _get_equality_components(self) -> tuple:
        return self._items
    
    def add(self, item: T) -> "ValueCollection[T]":
        """Return new collection with added item."""
        return ValueCollection(list(self._items) + [item])
    
    def remove(self, item: T) -> "ValueCollection[T]":
        """Return new collection without item."""
        items = [i for i in self._items if i != item]
        return ValueCollection(items)
    
    def filter(self, predicate: Callable[[T], bool]) -> "ValueCollection[T]":
        """Return filtered collection."""
        return ValueCollection([i for i in self._items if predicate(i)])
    
    def copy_with(self, **changes: Any) -> "ValueCollection[T]":
        return ValueCollection(list(self._items))


# Decorators
def value_object(cls: Type) -> Type:
    """
    Class decorator to mark as value object.
    
    Makes the class immutable (frozen dataclass) and adds equality.
    
    Example:
        @value_object
        class Color:
            red: int
            green: int
            blue: int
    """
    # Convert to frozen dataclass
    cls = dataclass(frozen=True)(cls)
    
    # Add value object methods
    def _get_equality_components(self) -> tuple:
        return tuple(getattr(self, f.name) for f in dataclass_fields(self))
    
    def copy_with(self, **changes):
        current = {f.name: getattr(self, f.name) for f in dataclass_fields(self)}
        current.update(changes)
        return self.__class__(**current)
    
    cls._get_equality_components = _get_equality_components
    cls.copy_with = copy_with
    
    return cls


def validated(validator: Callable[[Any], bool], message: str = "Validation failed"):
    """
    Decorator to add validation to value object.
    
    Example:
        @validated(lambda x: x.value > 0, "Value must be positive")
        @value_object
        class PositiveNumber:
            value: int
    """
    def decorator(cls: Type) -> Type:
        original_init = getattr(cls, '__post_init__', None)
        
        def __post_init__(self):
            if original_init:
                original_init(self)
            if not validator(self):
                raise ValidationError(message)
        
        cls.__post_init__ = __post_init__
        return cls
    
    return decorator


def immutable(cls: Type) -> Type:
    """
    Class decorator to make a class immutable.
    
    Example:
        @immutable
        class Config:
            host: str
            port: int
    """
    def __setattr__(self, name: str, value: Any):
        if hasattr(self, '_initialized'):
            raise ImmutabilityError(f"Cannot modify immutable attribute: {name}")
        object.__setattr__(self, name, value)
    
    def __delattr__(self, name: str):
        raise ImmutabilityError(f"Cannot delete immutable attribute: {name}")
    
    original_init = cls.__init__
    
    def __init__(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        object.__setattr__(self, '_initialized', True)
    
    cls.__setattr__ = __setattr__
    cls.__delattr__ = __delattr__
    cls.__init__ = __init__
    
    return cls


# Factory functions
def create_money(
    amount: Union[int, float, str, Decimal],
    currency: str = "USD",
) -> Money:
    """Create money value object."""
    return Money(Decimal(str(amount)), currency)


def create_email(address: str) -> EmailAddress:
    """Create email address value object."""
    return EmailAddress(address)


def create_phone(
    number: str,
    country_code: str = "+1",
) -> PhoneNumber:
    """Create phone number value object."""
    return PhoneNumber(number, country_code)


def create_address(
    street: str,
    city: str,
    state: str,
    postal_code: str,
    country: str = "USA",
    apartment: Optional[str] = None,
) -> Address:
    """Create address value object."""
    return Address(street, city, state, postal_code, country, apartment)


def create_date_range(
    start: date,
    end: date,
) -> DateRange:
    """Create date range value object."""
    return DateRange(start, end)


def create_percentage(value: Union[int, float, Decimal]) -> Percentage:
    """Create percentage value object."""
    return Percentage(Decimal(str(value)))


def create_quantity(value: int, unit: str = "pcs") -> Quantity:
    """Create quantity value object."""
    return Quantity(value, unit)


def create_coordinate(latitude: float, longitude: float) -> Coordinate:
    """Create coordinate value object."""
    return Coordinate(latitude, longitude)


def create_url(value: str) -> URL:
    """Create URL value object."""
    return URL(value)


def create_password(plain_text: str, min_length: int = 8) -> Password:
    """Create password value object."""
    return Password.create(plain_text, min_length)


def create_collection(items: List[T]) -> ValueCollection[T]:
    """Create value collection."""
    return ValueCollection(items)


__all__ = [
    # Exceptions
    "ValueObjectError",
    "ValidationError",
    "ImmutabilityError",
    # Base classes
    "ValueObject",
    "SimpleValueObject",
    "ValueCollection",
    # Common value objects
    "Money",
    "EmailAddress",
    "PhoneNumber",
    "Address",
    "DateRange",
    "Percentage",
    "Quantity",
    "Coordinate",
    "URL",
    "Password",
    # Decorators
    "value_object",
    "validated",
    "immutable",
    # Factory functions
    "create_money",
    "create_email",
    "create_phone",
    "create_address",
    "create_date_range",
    "create_percentage",
    "create_quantity",
    "create_coordinate",
    "create_url",
    "create_password",
    "create_collection",
]



"""
Enterprise Filter Module.

Provides request/response filtering, content transformation,
and data sanitization for agent operations.

Example:
    # Filter pipeline
    pipeline = FilterPipeline()
    pipeline.add(PiiFilter())
    pipeline.add(ContentFilter())
    
    filtered_request = await pipeline.filter_request(request)
    filtered_response = await pipeline.filter_response(response)
    
    # Decorators
    @filter_input(PiiFilter())
    async def process_data(data):
        ...
"""

from __future__ import annotations

import asyncio
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Pattern,
    Set,
    TypeVar,
    Union,
)
from datetime import datetime
from functools import wraps
from enum import Enum
import logging
import json
import hashlib

logger = logging.getLogger(__name__)

T = TypeVar('T')


class FilterError(Exception):
    """Filter error."""
    pass


class FilterBlockedError(FilterError):
    """Content was blocked by filter."""
    pass


class FilterAction(str, Enum):
    """Actions a filter can take."""
    PASS = "pass"
    MODIFY = "modify"
    BLOCK = "block"
    WARN = "warn"


@dataclass
class FilterResult:
    """Result of a filter operation."""
    action: FilterAction
    data: Any
    modified: bool = False
    blocked: bool = False
    warnings: List[str] = field(default_factory=list)
    modifications: List[str] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    @property
    def passed(self) -> bool:
        """Check if content passed the filter."""
        return self.action in (FilterAction.PASS, FilterAction.MODIFY, FilterAction.WARN)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "action": self.action.value,
            "modified": self.modified,
            "blocked": self.blocked,
            "warnings": self.warnings,
            "modifications": self.modifications,
            "metadata": self.metadata,
        }


class Filter(ABC, Generic[T]):
    """Abstract filter."""
    
    @property
    @abstractmethod
    def name(self) -> str:
        """Get filter name."""
        pass
    
    @abstractmethod
    async def filter(self, data: T) -> FilterResult:
        """Apply filter to data."""
        pass
    
    async def filter_request(self, data: T) -> FilterResult:
        """Filter incoming request."""
        return await self.filter(data)
    
    async def filter_response(self, data: T) -> FilterResult:
        """Filter outgoing response."""
        return await self.filter(data)


class PiiFilter(Filter[str]):
    """
    Filter for detecting and masking PII (Personally Identifiable Information).
    """
    
    PATTERNS = {
        "email": r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b',
        "phone": r'\b(?:\+?1[-.\s]?)?\(?[0-9]{3}\)?[-.\s]?[0-9]{3}[-.\s]?[0-9]{4}\b',
        "ssn": r'\b\d{3}[-.\s]?\d{2}[-.\s]?\d{4}\b',
        "credit_card": r'\b(?:\d{4}[-\s]?){3}\d{4}\b',
        "ip_address": r'\b(?:\d{1,3}\.){3}\d{1,3}\b',
    }
    
    def __init__(
        self,
        mask_char: str = "*",
        detect_only: bool = False,
        patterns: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize PII filter.
        
        Args:
            mask_char: Character to use for masking
            detect_only: Only detect, don't mask
            patterns: Custom regex patterns
        """
        self.mask_char = mask_char
        self.detect_only = detect_only
        self._patterns = {
            name: re.compile(pattern, re.IGNORECASE)
            for name, pattern in (patterns or self.PATTERNS).items()
        }
    
    @property
    def name(self) -> str:
        return "pii_filter"
    
    async def filter(self, data: str) -> FilterResult:
        """Filter PII from text."""
        if not isinstance(data, str):
            return FilterResult(action=FilterAction.PASS, data=data)
        
        modifications = []
        filtered_text = data
        
        for pii_type, pattern in self._patterns.items():
            matches = pattern.findall(data)
            
            if matches:
                modifications.append(f"Found {len(matches)} {pii_type} pattern(s)")
                
                if not self.detect_only:
                    filtered_text = pattern.sub(
                        lambda m: self.mask_char * len(m.group()),
                        filtered_text,
                    )
        
        if modifications:
            return FilterResult(
                action=FilterAction.MODIFY if not self.detect_only else FilterAction.WARN,
                data=filtered_text,
                modified=not self.detect_only and filtered_text != data,
                warnings=modifications if self.detect_only else [],
                modifications=modifications if not self.detect_only else [],
            )
        
        return FilterResult(action=FilterAction.PASS, data=data)


class ContentFilter(Filter[str]):
    """
    Filter for blocking or modifying inappropriate content.
    """
    
    def __init__(
        self,
        blocked_words: Optional[Set[str]] = None,
        replacement: str = "[FILTERED]",
        block_on_match: bool = False,
    ):
        """
        Initialize content filter.
        
        Args:
            blocked_words: Words to filter
            replacement: Replacement text
            block_on_match: Block entire content on match
        """
        self.blocked_words = blocked_words or set()
        self.replacement = replacement
        self.block_on_match = block_on_match
    
    @property
    def name(self) -> str:
        return "content_filter"
    
    async def filter(self, data: str) -> FilterResult:
        """Filter inappropriate content."""
        if not isinstance(data, str):
            return FilterResult(action=FilterAction.PASS, data=data)
        
        found_words = []
        filtered_text = data
        
        for word in self.blocked_words:
            pattern = re.compile(re.escape(word), re.IGNORECASE)
            if pattern.search(data):
                found_words.append(word)
                
                if self.block_on_match:
                    return FilterResult(
                        action=FilterAction.BLOCK,
                        data=data,
                        blocked=True,
                        modifications=[f"Blocked word found: {word}"],
                    )
                
                filtered_text = pattern.sub(self.replacement, filtered_text)
        
        if found_words:
            return FilterResult(
                action=FilterAction.MODIFY,
                data=filtered_text,
                modified=True,
                modifications=[f"Replaced {len(found_words)} blocked word(s)"],
            )
        
        return FilterResult(action=FilterAction.PASS, data=data)


class LengthFilter(Filter[str]):
    """
    Filter for limiting content length.
    """
    
    def __init__(
        self,
        max_length: int = 10000,
        truncate: bool = True,
        truncation_suffix: str = "...",
    ):
        """
        Initialize length filter.
        
        Args:
            max_length: Maximum allowed length
            truncate: Truncate if too long
            truncation_suffix: Suffix to add when truncating
        """
        self.max_length = max_length
        self.truncate = truncate
        self.truncation_suffix = truncation_suffix
    
    @property
    def name(self) -> str:
        return "length_filter"
    
    async def filter(self, data: str) -> FilterResult:
        """Filter by length."""
        if not isinstance(data, str):
            return FilterResult(action=FilterAction.PASS, data=data)
        
        if len(data) <= self.max_length:
            return FilterResult(action=FilterAction.PASS, data=data)
        
        if self.truncate:
            truncated = data[:self.max_length - len(self.truncation_suffix)] + self.truncation_suffix
            return FilterResult(
                action=FilterAction.MODIFY,
                data=truncated,
                modified=True,
                modifications=[f"Truncated from {len(data)} to {len(truncated)} chars"],
            )
        
        return FilterResult(
            action=FilterAction.BLOCK,
            data=data,
            blocked=True,
            modifications=[f"Content length {len(data)} exceeds limit {self.max_length}"],
        )


class JsonFilter(Filter[str]):
    """
    Filter for validating and sanitizing JSON.
    """
    
    def __init__(
        self,
        allowed_keys: Optional[Set[str]] = None,
        blocked_keys: Optional[Set[str]] = None,
        max_depth: int = 10,
    ):
        """
        Initialize JSON filter.
        
        Args:
            allowed_keys: Only allow these keys
            blocked_keys: Block these keys
            max_depth: Maximum nesting depth
        """
        self.allowed_keys = allowed_keys
        self.blocked_keys = blocked_keys or set()
        self.max_depth = max_depth
    
    @property
    def name(self) -> str:
        return "json_filter"
    
    def _filter_dict(
        self,
        data: Dict[str, Any],
        depth: int = 0,
    ) -> Dict[str, Any]:
        """Recursively filter dictionary."""
        if depth > self.max_depth:
            return {}
        
        filtered = {}
        
        for key, value in data.items():
            # Skip blocked keys
            if key in self.blocked_keys:
                continue
            
            # Check allowed keys
            if self.allowed_keys and key not in self.allowed_keys:
                continue
            
            # Recurse into nested dicts
            if isinstance(value, dict):
                filtered[key] = self._filter_dict(value, depth + 1)
            elif isinstance(value, list):
                filtered[key] = [
                    self._filter_dict(item, depth + 1) if isinstance(item, dict) else item
                    for item in value
                ]
            else:
                filtered[key] = value
        
        return filtered
    
    async def filter(self, data: str) -> FilterResult:
        """Filter JSON content."""
        try:
            parsed = json.loads(data) if isinstance(data, str) else data
        except json.JSONDecodeError as e:
            return FilterResult(
                action=FilterAction.BLOCK,
                data=data,
                blocked=True,
                modifications=[f"Invalid JSON: {e}"],
            )
        
        if isinstance(parsed, dict):
            filtered = self._filter_dict(parsed)
            filtered_str = json.dumps(filtered)
            
            if filtered != parsed:
                return FilterResult(
                    action=FilterAction.MODIFY,
                    data=filtered_str,
                    modified=True,
                    modifications=["Removed blocked/unauthorized keys"],
                )
            
            return FilterResult(action=FilterAction.PASS, data=data)
        
        return FilterResult(action=FilterAction.PASS, data=data)


class RegexFilter(Filter[str]):
    """
    Filter using regular expressions.
    """
    
    def __init__(
        self,
        patterns: List[Dict[str, Any]],
    ):
        """
        Initialize regex filter.
        
        Args:
            patterns: List of {"pattern": str, "action": str, "replacement": str}
        """
        self._patterns = []
        
        for p in patterns:
            self._patterns.append({
                "regex": re.compile(p["pattern"], re.IGNORECASE),
                "action": FilterAction(p.get("action", "modify")),
                "replacement": p.get("replacement", ""),
            })
    
    @property
    def name(self) -> str:
        return "regex_filter"
    
    async def filter(self, data: str) -> FilterResult:
        """Apply regex filters."""
        if not isinstance(data, str):
            return FilterResult(action=FilterAction.PASS, data=data)
        
        filtered_text = data
        modifications = []
        blocked = False
        
        for p in self._patterns:
            if p["regex"].search(data):
                if p["action"] == FilterAction.BLOCK:
                    blocked = True
                    modifications.append("Blocked by regex pattern")
                    break
                elif p["action"] == FilterAction.MODIFY:
                    filtered_text = p["regex"].sub(p["replacement"], filtered_text)
                    modifications.append("Modified by regex pattern")
                elif p["action"] == FilterAction.WARN:
                    modifications.append("Warning: regex pattern matched")
        
        if blocked:
            return FilterResult(
                action=FilterAction.BLOCK,
                data=data,
                blocked=True,
                modifications=modifications,
            )
        
        if modifications:
            return FilterResult(
                action=FilterAction.MODIFY if filtered_text != data else FilterAction.WARN,
                data=filtered_text,
                modified=filtered_text != data,
                warnings=[m for m in modifications if "Warning" in m],
                modifications=[m for m in modifications if "Warning" not in m],
            )
        
        return FilterResult(action=FilterAction.PASS, data=data)


class CompositeFilter(Filter[T]):
    """
    Composite filter that applies multiple filters.
    """
    
    def __init__(
        self,
        filters: Optional[List[Filter]] = None,
        stop_on_block: bool = True,
    ):
        """
        Initialize composite filter.
        
        Args:
            filters: List of filters to apply
            stop_on_block: Stop processing if any filter blocks
        """
        self._filters = list(filters) if filters else []
        self._stop_on_block = stop_on_block
    
    @property
    def name(self) -> str:
        return "composite_filter"
    
    def add(self, filter_: Filter) -> 'CompositeFilter':
        """Add a filter."""
        self._filters.append(filter_)
        return self
    
    async def filter(self, data: T) -> FilterResult:
        """Apply all filters."""
        current_data = data
        all_modifications = []
        all_warnings = []
        was_modified = False
        
        for f in self._filters:
            result = await f.filter(current_data)
            
            if result.blocked:
                return FilterResult(
                    action=FilterAction.BLOCK,
                    data=current_data,
                    blocked=True,
                    modifications=all_modifications + result.modifications,
                    warnings=all_warnings + result.warnings,
                    metadata={"blocked_by": f.name},
                )
            
            if result.modified:
                current_data = result.data
                was_modified = True
            
            all_modifications.extend(result.modifications)
            all_warnings.extend(result.warnings)
        
        return FilterResult(
            action=FilterAction.MODIFY if was_modified else FilterAction.PASS,
            data=current_data,
            modified=was_modified,
            modifications=all_modifications,
            warnings=all_warnings,
        )


class FilterPipeline:
    """
    Pipeline for filtering requests and responses.
    """
    
    def __init__(self):
        self._request_filters: List[Filter] = []
        self._response_filters: List[Filter] = []
    
    def add_request_filter(self, filter_: Filter) -> 'FilterPipeline':
        """Add a request filter."""
        self._request_filters.append(filter_)
        return self
    
    def add_response_filter(self, filter_: Filter) -> 'FilterPipeline':
        """Add a response filter."""
        self._response_filters.append(filter_)
        return self
    
    def add(self, filter_: Filter) -> 'FilterPipeline':
        """Add filter for both request and response."""
        self._request_filters.append(filter_)
        self._response_filters.append(filter_)
        return self
    
    async def filter_request(self, data: Any) -> FilterResult:
        """Filter incoming request."""
        composite = CompositeFilter(self._request_filters)
        return await composite.filter(data)
    
    async def filter_response(self, data: Any) -> FilterResult:
        """Filter outgoing response."""
        composite = CompositeFilter(self._response_filters)
        return await composite.filter(data)


class FilterMiddleware:
    """
    Middleware for agent filter pipeline.
    """
    
    def __init__(self, pipeline: FilterPipeline):
        self.pipeline = pipeline
    
    async def __call__(
        self,
        request: Any,
        next_handler: Callable,
    ) -> Any:
        """Process request through filter pipeline."""
        # Filter request
        request_result = await self.pipeline.filter_request(request)
        
        if request_result.blocked:
            raise FilterBlockedError(
                f"Request blocked: {request_result.modifications}"
            )
        
        # Call next handler
        response = await next_handler(request_result.data)
        
        # Filter response
        response_result = await self.pipeline.filter_response(response)
        
        if response_result.blocked:
            raise FilterBlockedError(
                f"Response blocked: {response_result.modifications}"
            )
        
        return response_result.data


def filter_input(
    *filters: Filter,
) -> Callable:
    """
    Decorator to filter function input.
    
    Example:
        @filter_input(PiiFilter(), ContentFilter())
        async def process(data):
            ...
    """
    composite = CompositeFilter(list(filters))
    
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            # Filter first argument
            if args:
                result = await composite.filter(args[0])
                if result.blocked:
                    raise FilterBlockedError(f"Input blocked: {result.modifications}")
                args = (result.data,) + args[1:]
            
            return await func(*args, **kwargs)
        
        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            if args:
                result = asyncio.run(composite.filter(args[0]))
                if result.blocked:
                    raise FilterBlockedError(f"Input blocked: {result.modifications}")
                args = (result.data,) + args[1:]
            
            return func(*args, **kwargs)
        
        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
    
    return decorator


def filter_output(
    *filters: Filter,
) -> Callable:
    """
    Decorator to filter function output.
    
    Example:
        @filter_output(PiiFilter())
        async def generate():
            return "some data with email@example.com"
    """
    composite = CompositeFilter(list(filters))
    
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            result = await func(*args, **kwargs)
            
            filter_result = await composite.filter(result)
            if filter_result.blocked:
                raise FilterBlockedError(f"Output blocked: {filter_result.modifications}")
            
            return filter_result.data
        
        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            result = func(*args, **kwargs)
            
            filter_result = asyncio.run(composite.filter(result))
            if filter_result.blocked:
                raise FilterBlockedError(f"Output blocked: {filter_result.modifications}")
            
            return filter_result.data
        
        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
    
    return decorator


__all__ = [
    # Exceptions
    "FilterError",
    "FilterBlockedError",
    # Enums
    "FilterAction",
    # Data classes
    "FilterResult",
    # Filters
    "Filter",
    "PiiFilter",
    "ContentFilter",
    "LengthFilter",
    "JsonFilter",
    "RegexFilter",
    "CompositeFilter",
    # Pipeline
    "FilterPipeline",
    "FilterMiddleware",
    # Decorators
    "filter_input",
    "filter_output",
]



"""
Circuit Breaker.

Circuit breaker pattern to prevent cascading failures.
Thread-safe implementation with proper exception re-raising.
"""

from __future__ import annotations

import logging
import threading
import time
from typing import Any, Callable

from ..exceptions import CircuitBreakerOpenError

logger = logging.getLogger(__name__)


class CircuitBreaker:
    """Thread-safe circuit breaker to prevent cascading failures."""

    __slots__ = (
        "failure_threshold",
        "recovery_timeout",
        "failure_count",
        "last_failure_time",
        "state",
        "_lock",
    )

    def __init__(
        self,
        failure_threshold: int = 5,
        recovery_timeout: int = 60,
    ):
        self.failure_threshold = failure_threshold
        self.recovery_timeout = recovery_timeout
        self.failure_count = 0
        self.last_failure_time: float | None = None
        self.state = "closed"  # closed, open, half-open
        self._lock = threading.Lock()

    def call(self, func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        """Execute function with circuit breaker protection."""
        with self._lock:
            if self.state == "open":
                if (
                    self.last_failure_time
                    and time.time() - self.last_failure_time > self.recovery_timeout
                ):
                    self.state = "half-open"
                    self.failure_count = 0
                    logger.info("Circuit breaker transitioning to half-open")
                else:
                    raise CircuitBreakerOpenError(
                        recovery_timeout=self.recovery_timeout
                    )

        try:
            result = func(*args, **kwargs)

            with self._lock:
                if self.state == "half-open":
                    self.state = "closed"
                    self.failure_count = 0
                    logger.info("Circuit breaker closed after successful call")

            return result

        except Exception:
            with self._lock:
                self.failure_count += 1
                self.last_failure_time = time.time()

                if self.failure_count >= self.failure_threshold:
                    self.state = "open"
                    logger.warning(
                        "Circuit breaker opened after %d failures",
                        self.failure_count,
                    )

            raise  # preserve original traceback

    def reset(self) -> None:
        """Manually reset the circuit breaker."""
        with self._lock:
            self.state = "closed"
            self.failure_count = 0
            self.last_failure_time = None
            logger.info("Circuit breaker manually reset")


__all__ = ['CircuitBreaker']



"""
Enterprise Inventory Manager Module.

Stock tracking, SKU management, warehouses,
reorder alerts, and inventory operations.

Example:
    # Create inventory manager
    inventory = create_inventory_manager()
    
    # Add product
    product = await inventory.add_product(
        sku="WIDGET-001",
        name="Blue Widget",
        category="widgets",
    )
    
    # Stock product
    await inventory.stock(
        sku="WIDGET-001",
        warehouse_id="wh_main",
        quantity=100,
        unit_cost=5.00,
    )
    
    # Reserve for order
    reservation = await inventory.reserve(
        sku="WIDGET-001",
        quantity=5,
        order_id="order_123",
    )
    
    # Check stock
    stock = await inventory.get_stock("WIDGET-001")
"""

from __future__ import annotations

import asyncio
import json
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class InventoryError(Exception):
    """Inventory error."""
    pass


class ProductNotFoundError(InventoryError):
    """Product not found."""
    pass


class InsufficientStockError(InventoryError):
    """Insufficient stock."""
    pass


class StockStatus(str, Enum):
    """Stock status."""
    IN_STOCK = "in_stock"
    LOW_STOCK = "low_stock"
    OUT_OF_STOCK = "out_of_stock"
    DISCONTINUED = "discontinued"


class MovementType(str, Enum):
    """Stock movement type."""
    PURCHASE = "purchase"
    SALE = "sale"
    RETURN = "return"
    ADJUSTMENT = "adjustment"
    TRANSFER = "transfer"
    DAMAGE = "damage"
    EXPIRED = "expired"


class ReorderStrategy(str, Enum):
    """Reorder strategy."""
    FIXED_QUANTITY = "fixed_quantity"
    ECONOMIC_ORDER = "economic_order"
    MIN_MAX = "min_max"


@dataclass
class Product:
    """Product definition."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    sku: str = ""
    name: str = ""
    description: str = ""
    category: str = ""
    brand: str = ""
    unit: str = "each"  # each, kg, liter, etc.
    weight: float = 0.0  # kg
    dimensions: Dict[str, float] = field(default_factory=dict)  # length, width, height
    barcode: str = ""
    reorder_point: int = 10
    reorder_quantity: int = 50
    max_stock: int = 1000
    lead_time_days: int = 7
    is_active: bool = True
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Warehouse:
    """Warehouse location."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    code: str = ""
    address: str = ""
    city: str = ""
    country: str = ""
    is_primary: bool = False
    is_active: bool = True
    capacity: int = 0
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class StockLevel:
    """Stock level at warehouse."""
    product_id: str = ""
    sku: str = ""
    warehouse_id: str = ""
    quantity: int = 0
    reserved: int = 0
    unit_cost: float = 0.0
    bin_location: str = ""
    lot_number: str = ""
    expiry_date: Optional[datetime] = None
    last_counted: Optional[datetime] = None
    updated_at: datetime = field(default_factory=datetime.utcnow)
    
    @property
    def available(self) -> int:
        """Available quantity."""
        return max(0, self.quantity - self.reserved)
    
    @property
    def total_value(self) -> float:
        """Total value."""
        return self.quantity * self.unit_cost


@dataclass
class StockMovement:
    """Stock movement record."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    product_id: str = ""
    sku: str = ""
    warehouse_id: str = ""
    type: MovementType = MovementType.ADJUSTMENT
    quantity: int = 0
    unit_cost: float = 0.0
    reference_id: str = ""  # order_id, transfer_id, etc.
    reason: str = ""
    performed_by: str = ""
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Reservation:
    """Stock reservation."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    product_id: str = ""
    sku: str = ""
    warehouse_id: str = ""
    quantity: int = 0
    order_id: str = ""
    expires_at: Optional[datetime] = None
    fulfilled: bool = False
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class ReorderAlert:
    """Reorder alert."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    product_id: str = ""
    sku: str = ""
    warehouse_id: str = ""
    current_quantity: int = 0
    reorder_point: int = 0
    suggested_quantity: int = 0
    priority: str = "normal"  # low, normal, high, critical
    acknowledged: bool = False
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class InventoryStats:
    """Inventory statistics."""
    total_products: int = 0
    total_warehouses: int = 0
    total_stock_value: float = 0.0
    low_stock_items: int = 0
    out_of_stock_items: int = 0


# Product store
class ProductStore(ABC):
    """Product storage."""
    
    @abstractmethod
    async def save(self, product: Product) -> None:
        pass
    
    @abstractmethod
    async def get(self, product_id: str) -> Optional[Product]:
        pass
    
    @abstractmethod
    async def get_by_sku(self, sku: str) -> Optional[Product]:
        pass
    
    @abstractmethod
    async def list(
        self,
        category: Optional[str] = None,
        active_only: bool = True,
    ) -> List[Product]:
        pass


class InMemoryProductStore(ProductStore):
    """In-memory product store."""
    
    def __init__(self):
        self._products: Dict[str, Product] = {}
        self._sku_index: Dict[str, str] = {}
    
    async def save(self, product: Product) -> None:
        self._products[product.id] = product
        self._sku_index[product.sku] = product.id
    
    async def get(self, product_id: str) -> Optional[Product]:
        return self._products.get(product_id)
    
    async def get_by_sku(self, sku: str) -> Optional[Product]:
        product_id = self._sku_index.get(sku)
        return self._products.get(product_id) if product_id else None
    
    async def list(
        self,
        category: Optional[str] = None,
        active_only: bool = True,
    ) -> List[Product]:
        products = list(self._products.values())
        
        if category:
            products = [p for p in products if p.category == category]
        if active_only:
            products = [p for p in products if p.is_active]
        
        return products


# Stock store
class StockStore(ABC):
    """Stock storage."""
    
    @abstractmethod
    async def save(self, stock: StockLevel) -> None:
        pass
    
    @abstractmethod
    async def get(
        self,
        product_id: str,
        warehouse_id: str,
    ) -> Optional[StockLevel]:
        pass
    
    @abstractmethod
    async def list(
        self,
        product_id: Optional[str] = None,
        warehouse_id: Optional[str] = None,
    ) -> List[StockLevel]:
        pass


class InMemoryStockStore(StockStore):
    """In-memory stock store."""
    
    def __init__(self):
        self._stock: Dict[str, StockLevel] = {}  # key: product_id:warehouse_id
    
    def _key(self, product_id: str, warehouse_id: str) -> str:
        return f"{product_id}:{warehouse_id}"
    
    async def save(self, stock: StockLevel) -> None:
        key = self._key(stock.product_id, stock.warehouse_id)
        stock.updated_at = datetime.utcnow()
        self._stock[key] = stock
    
    async def get(
        self,
        product_id: str,
        warehouse_id: str,
    ) -> Optional[StockLevel]:
        key = self._key(product_id, warehouse_id)
        return self._stock.get(key)
    
    async def list(
        self,
        product_id: Optional[str] = None,
        warehouse_id: Optional[str] = None,
    ) -> List[StockLevel]:
        results = list(self._stock.values())
        
        if product_id:
            results = [s for s in results if s.product_id == product_id]
        if warehouse_id:
            results = [s for s in results if s.warehouse_id == warehouse_id]
        
        return results


# Inventory manager
class InventoryManager:
    """Inventory manager."""
    
    def __init__(
        self,
        product_store: Optional[ProductStore] = None,
        stock_store: Optional[StockStore] = None,
    ):
        self._products = product_store or InMemoryProductStore()
        self._stock = stock_store or InMemoryStockStore()
        self._warehouses: Dict[str, Warehouse] = {}
        self._movements: List[StockMovement] = []
        self._reservations: Dict[str, Reservation] = {}
        self._alerts: List[ReorderAlert] = []
        self._stats = InventoryStats()
    
    # Product management
    async def add_product(
        self,
        sku: str,
        name: str,
        category: str = "",
        reorder_point: int = 10,
        reorder_quantity: int = 50,
        **kwargs,
    ) -> Product:
        """Add product."""
        product = Product(
            sku=sku,
            name=name,
            category=category,
            reorder_point=reorder_point,
            reorder_quantity=reorder_quantity,
            **kwargs,
        )
        await self._products.save(product)
        self._stats.total_products += 1
        
        logger.info(f"Product added: {sku}")
        
        return product
    
    async def get_product(self, sku: str) -> Optional[Product]:
        """Get product by SKU."""
        return await self._products.get_by_sku(sku)
    
    async def list_products(
        self,
        category: Optional[str] = None,
    ) -> List[Product]:
        """List products."""
        return await self._products.list(category=category)
    
    # Warehouse management
    async def add_warehouse(
        self,
        name: str,
        code: str,
        **kwargs,
    ) -> Warehouse:
        """Add warehouse."""
        warehouse = Warehouse(
            name=name,
            code=code,
            **kwargs,
        )
        self._warehouses[warehouse.id] = warehouse
        self._stats.total_warehouses += 1
        
        logger.info(f"Warehouse added: {name}")
        
        return warehouse
    
    async def get_warehouse(self, warehouse_id: str) -> Optional[Warehouse]:
        """Get warehouse."""
        return self._warehouses.get(warehouse_id)
    
    async def list_warehouses(self) -> List[Warehouse]:
        """List warehouses."""
        return [w for w in self._warehouses.values() if w.is_active]
    
    # Stock operations
    async def stock(
        self,
        sku: str,
        warehouse_id: str,
        quantity: int,
        unit_cost: float = 0.0,
        lot_number: str = "",
        expiry_date: Optional[datetime] = None,
        performed_by: str = "",
    ) -> StockLevel:
        """Add stock."""
        product = await self._products.get_by_sku(sku)
        if not product:
            raise ProductNotFoundError(f"Product not found: {sku}")
        
        # Get or create stock level
        stock = await self._stock.get(product.id, warehouse_id)
        if not stock:
            stock = StockLevel(
                product_id=product.id,
                sku=sku,
                warehouse_id=warehouse_id,
            )
        
        stock.quantity += quantity
        stock.unit_cost = unit_cost or stock.unit_cost
        stock.lot_number = lot_number or stock.lot_number
        stock.expiry_date = expiry_date or stock.expiry_date
        
        await self._stock.save(stock)
        
        # Record movement
        movement = StockMovement(
            product_id=product.id,
            sku=sku,
            warehouse_id=warehouse_id,
            type=MovementType.PURCHASE,
            quantity=quantity,
            unit_cost=unit_cost,
            performed_by=performed_by,
        )
        self._movements.append(movement)
        
        # Update stats
        self._stats.total_stock_value += quantity * unit_cost
        
        logger.info(f"Stock added: {sku} x {quantity}")
        
        return stock
    
    async def get_stock(
        self,
        sku: str,
        warehouse_id: Optional[str] = None,
    ) -> List[StockLevel]:
        """Get stock levels."""
        product = await self._products.get_by_sku(sku)
        if not product:
            return []
        
        stocks = await self._stock.list(
            product_id=product.id,
            warehouse_id=warehouse_id,
        )
        
        return stocks
    
    async def get_total_stock(self, sku: str) -> int:
        """Get total stock across all warehouses."""
        stocks = await self.get_stock(sku)
        return sum(s.available for s in stocks)
    
    async def get_stock_status(self, sku: str) -> StockStatus:
        """Get stock status."""
        product = await self._products.get_by_sku(sku)
        if not product:
            return StockStatus.OUT_OF_STOCK
        
        if not product.is_active:
            return StockStatus.DISCONTINUED
        
        total = await self.get_total_stock(sku)
        
        if total == 0:
            return StockStatus.OUT_OF_STOCK
        elif total <= product.reorder_point:
            return StockStatus.LOW_STOCK
        else:
            return StockStatus.IN_STOCK
    
    # Reservation
    async def reserve(
        self,
        sku: str,
        quantity: int,
        order_id: str,
        warehouse_id: Optional[str] = None,
        expires_in: int = 1800,  # seconds
    ) -> Reservation:
        """Reserve stock."""
        product = await self._products.get_by_sku(sku)
        if not product:
            raise ProductNotFoundError(f"Product not found: {sku}")
        
        # Find stock to reserve
        stocks = await self.get_stock(sku, warehouse_id)
        
        remaining = quantity
        reservations: List[Reservation] = []
        
        for stock in stocks:
            if remaining <= 0:
                break
            
            available = stock.available
            if available <= 0:
                continue
            
            reserve_qty = min(remaining, available)
            stock.reserved += reserve_qty
            await self._stock.save(stock)
            
            reservation = Reservation(
                product_id=product.id,
                sku=sku,
                warehouse_id=stock.warehouse_id,
                quantity=reserve_qty,
                order_id=order_id,
                expires_at=datetime.utcnow() + timedelta(seconds=expires_in),
            )
            self._reservations[reservation.id] = reservation
            reservations.append(reservation)
            
            remaining -= reserve_qty
        
        if remaining > 0:
            # Rollback reservations
            for r in reservations:
                await self.release_reservation(r.id)
            raise InsufficientStockError(
                f"Insufficient stock for {sku}. Available: {quantity - remaining}"
            )
        
        # Return first reservation (main one)
        return reservations[0] if reservations else Reservation()
    
    async def release_reservation(self, reservation_id: str) -> bool:
        """Release reservation."""
        reservation = self._reservations.get(reservation_id)
        if not reservation:
            return False
        
        stock = await self._stock.get(
            reservation.product_id,
            reservation.warehouse_id,
        )
        if stock:
            stock.reserved -= reservation.quantity
            await self._stock.save(stock)
        
        del self._reservations[reservation_id]
        
        return True
    
    async def fulfill_reservation(self, reservation_id: str) -> bool:
        """Fulfill reservation (convert to sale)."""
        reservation = self._reservations.get(reservation_id)
        if not reservation:
            return False
        
        stock = await self._stock.get(
            reservation.product_id,
            reservation.warehouse_id,
        )
        if stock:
            stock.quantity -= reservation.quantity
            stock.reserved -= reservation.quantity
            await self._stock.save(stock)
            
            # Record movement
            movement = StockMovement(
                product_id=reservation.product_id,
                sku=reservation.sku,
                warehouse_id=reservation.warehouse_id,
                type=MovementType.SALE,
                quantity=-reservation.quantity,
                reference_id=reservation.order_id,
            )
            self._movements.append(movement)
        
        reservation.fulfilled = True
        del self._reservations[reservation_id]
        
        # Check reorder
        await self._check_reorder(reservation.sku, reservation.warehouse_id)
        
        return True
    
    # Transfer
    async def transfer(
        self,
        sku: str,
        from_warehouse: str,
        to_warehouse: str,
        quantity: int,
        performed_by: str = "",
    ) -> Tuple[StockLevel, StockLevel]:
        """Transfer stock between warehouses."""
        product = await self._products.get_by_sku(sku)
        if not product:
            raise ProductNotFoundError(f"Product not found: {sku}")
        
        # Get source stock
        source = await self._stock.get(product.id, from_warehouse)
        if not source or source.available < quantity:
            raise InsufficientStockError(
                f"Insufficient stock in source warehouse"
            )
        
        # Get or create destination stock
        dest = await self._stock.get(product.id, to_warehouse)
        if not dest:
            dest = StockLevel(
                product_id=product.id,
                sku=sku,
                warehouse_id=to_warehouse,
                unit_cost=source.unit_cost,
            )
        
        # Transfer
        source.quantity -= quantity
        dest.quantity += quantity
        
        await self._stock.save(source)
        await self._stock.save(dest)
        
        # Record movements
        self._movements.append(StockMovement(
            product_id=product.id,
            sku=sku,
            warehouse_id=from_warehouse,
            type=MovementType.TRANSFER,
            quantity=-quantity,
            reference_id=to_warehouse,
            performed_by=performed_by,
        ))
        self._movements.append(StockMovement(
            product_id=product.id,
            sku=sku,
            warehouse_id=to_warehouse,
            type=MovementType.TRANSFER,
            quantity=quantity,
            reference_id=from_warehouse,
            performed_by=performed_by,
        ))
        
        logger.info(f"Transferred {quantity} x {sku} from {from_warehouse} to {to_warehouse}")
        
        return source, dest
    
    # Adjustment
    async def adjust(
        self,
        sku: str,
        warehouse_id: str,
        quantity: int,  # Can be negative
        reason: str = "",
        performed_by: str = "",
    ) -> StockLevel:
        """Adjust stock."""
        product = await self._products.get_by_sku(sku)
        if not product:
            raise ProductNotFoundError(f"Product not found: {sku}")
        
        stock = await self._stock.get(product.id, warehouse_id)
        if not stock:
            stock = StockLevel(
                product_id=product.id,
                sku=sku,
                warehouse_id=warehouse_id,
            )
        
        stock.quantity += quantity
        if stock.quantity < 0:
            stock.quantity = 0
        
        await self._stock.save(stock)
        
        # Record movement
        self._movements.append(StockMovement(
            product_id=product.id,
            sku=sku,
            warehouse_id=warehouse_id,
            type=MovementType.ADJUSTMENT,
            quantity=quantity,
            reason=reason,
            performed_by=performed_by,
        ))
        
        logger.info(f"Stock adjusted: {sku} by {quantity}")
        
        return stock
    
    # Reorder
    async def _check_reorder(
        self,
        sku: str,
        warehouse_id: str,
    ) -> Optional[ReorderAlert]:
        """Check if reorder is needed."""
        product = await self._products.get_by_sku(sku)
        if not product:
            return None
        
        stock = await self._stock.get(product.id, warehouse_id)
        if not stock:
            return None
        
        if stock.quantity <= product.reorder_point:
            alert = ReorderAlert(
                product_id=product.id,
                sku=sku,
                warehouse_id=warehouse_id,
                current_quantity=stock.quantity,
                reorder_point=product.reorder_point,
                suggested_quantity=product.reorder_quantity,
                priority="high" if stock.quantity == 0 else "normal",
            )
            self._alerts.append(alert)
            self._stats.low_stock_items += 1
            
            if stock.quantity == 0:
                self._stats.out_of_stock_items += 1
            
            logger.warning(f"Reorder alert: {sku} at {stock.quantity}")
            
            return alert
        
        return None
    
    async def get_reorder_alerts(
        self,
        acknowledged: bool = False,
    ) -> List[ReorderAlert]:
        """Get reorder alerts."""
        return [
            a for a in self._alerts
            if a.acknowledged == acknowledged
        ]
    
    async def acknowledge_alert(self, alert_id: str) -> bool:
        """Acknowledge alert."""
        for alert in self._alerts:
            if alert.id == alert_id:
                alert.acknowledged = True
                return True
        return False
    
    # Movements
    async def get_movements(
        self,
        sku: Optional[str] = None,
        warehouse_id: Optional[str] = None,
        type: Optional[MovementType] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[StockMovement]:
        """Get stock movements."""
        movements = self._movements
        
        if sku:
            movements = [m for m in movements if m.sku == sku]
        if warehouse_id:
            movements = [m for m in movements if m.warehouse_id == warehouse_id]
        if type:
            movements = [m for m in movements if m.type == type]
        if start:
            movements = [m for m in movements if m.created_at >= start]
        if end:
            movements = [m for m in movements if m.created_at <= end]
        
        return sorted(movements, key=lambda m: m.created_at, reverse=True)
    
    # Stats
    def get_stats(self) -> InventoryStats:
        """Get statistics."""
        return self._stats


# Factory functions
def create_inventory_manager() -> InventoryManager:
    """Create inventory manager."""
    return InventoryManager()


def create_product(
    sku: str,
    name: str,
    **kwargs,
) -> Product:
    """Create product."""
    return Product(sku=sku, name=name, **kwargs)


def create_warehouse(
    name: str,
    code: str,
    **kwargs,
) -> Warehouse:
    """Create warehouse."""
    return Warehouse(name=name, code=code, **kwargs)


__all__ = [
    # Exceptions
    "InventoryError",
    "ProductNotFoundError",
    "InsufficientStockError",
    # Enums
    "StockStatus",
    "MovementType",
    "ReorderStrategy",
    # Data classes
    "Product",
    "Warehouse",
    "StockLevel",
    "StockMovement",
    "Reservation",
    "ReorderAlert",
    "InventoryStats",
    # Stores
    "ProductStore",
    "InMemoryProductStore",
    "StockStore",
    "InMemoryStockStore",
    # Manager
    "InventoryManager",
    # Factory functions
    "create_inventory_manager",
    "create_product",
    "create_warehouse",
]



"""
Enterprise Validation Module.

Provides input/output validation, schema validation, and sanitization
for secure and reliable agent interactions.

Example:
    # Schema validation
    @validate_input(UserSchema)
    @validate_output(ResponseSchema)
    async def process_user(data: dict) -> dict:
        return {"id": data["id"], "status": "processed"}
    
    # Field validation
    validator = Validator()
    validator.add_rule("email", EmailRule())
    validator.add_rule("age", RangeRule(min=0, max=150))
    
    result = validator.validate(data)
"""

from __future__ import annotations

import re
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Pattern,
    Set,
    Type,
    TypeVar,
    Union,
)
from functools import wraps
import logging

logger = logging.getLogger(__name__)

T = TypeVar("T")


class ValidationError(Exception):
    """Validation error with details."""
    
    def __init__(
        self,
        message: str,
        field: Optional[str] = None,
        value: Any = None,
        errors: Optional[List['ValidationError']] = None,
    ):
        super().__init__(message)
        self.field = field
        self.value = value
        self.errors = errors or []
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        result = {"message": str(self)}
        if self.field:
            result["field"] = self.field
        if self.errors:
            result["errors"] = [e.to_dict() for e in self.errors]
        return result


@dataclass
class ValidationResult:
    """Result of validation operation."""
    valid: bool
    errors: List[ValidationError] = field(default_factory=list)
    warnings: List[str] = field(default_factory=list)
    sanitized_data: Optional[Any] = None
    
    def __bool__(self) -> bool:
        return self.valid
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "valid": self.valid,
            "errors": [e.to_dict() for e in self.errors],
            "warnings": self.warnings,
        }


class ValidationRule(ABC):
    """Abstract validation rule."""
    
    @abstractmethod
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        """Validate a value."""
        pass
    
    def sanitize(self, value: Any) -> Any:
        """Sanitize value (optional)."""
        return value


class RequiredRule(ValidationRule):
    """Validates that a value is present."""
    
    def __init__(self, message: str = "Field is required"):
        self.message = message
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None or value == "" or (isinstance(value, (list, dict)) and len(value) == 0):
            return ValidationResult(
                valid=False,
                errors=[ValidationError(self.message, field_name, value)],
            )
        return ValidationResult(valid=True, sanitized_data=value)


class TypeRule(ValidationRule):
    """Validates value type."""
    
    def __init__(
        self,
        expected_type: Type,
        coerce: bool = False,
        message: Optional[str] = None,
    ):
        self.expected_type = expected_type
        self.coerce = coerce
        self.message = message
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None:
            return ValidationResult(valid=True, sanitized_data=value)
        
        if isinstance(value, self.expected_type):
            return ValidationResult(valid=True, sanitized_data=value)
        
        if self.coerce:
            try:
                coerced = self.expected_type(value)
                return ValidationResult(valid=True, sanitized_data=coerced)
            except (ValueError, TypeError):
                pass
        
        message = self.message or f"Expected {self.expected_type.__name__}, got {type(value).__name__}"
        return ValidationResult(
            valid=False,
            errors=[ValidationError(message, field_name, value)],
        )


class StringRule(ValidationRule):
    """Validates string values."""
    
    def __init__(
        self,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[Union[str, Pattern]] = None,
        strip: bool = True,
        lowercase: bool = False,
        uppercase: bool = False,
    ):
        self.min_length = min_length
        self.max_length = max_length
        self.pattern = re.compile(pattern) if isinstance(pattern, str) else pattern
        self.strip = strip
        self.lowercase = lowercase
        self.uppercase = uppercase
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None:
            return ValidationResult(valid=True, sanitized_data=value)
        
        if not isinstance(value, str):
            return ValidationResult(
                valid=False,
                errors=[ValidationError(f"Expected string, got {type(value).__name__}", field_name, value)],
            )
        
        # Sanitize
        sanitized = self.sanitize(value)
        errors = []
        
        if self.min_length and len(sanitized) < self.min_length:
            errors.append(ValidationError(
                f"Minimum length is {self.min_length}",
                field_name,
                value,
            ))
        
        if self.max_length and len(sanitized) > self.max_length:
            errors.append(ValidationError(
                f"Maximum length is {self.max_length}",
                field_name,
                value,
            ))
        
        if self.pattern and not self.pattern.match(sanitized):
            errors.append(ValidationError(
                f"Value does not match required pattern",
                field_name,
                value,
            ))
        
        return ValidationResult(
            valid=len(errors) == 0,
            errors=errors,
            sanitized_data=sanitized,
        )
    
    def sanitize(self, value: Any) -> str:
        if not isinstance(value, str):
            return value
        
        result = value
        if self.strip:
            result = result.strip()
        if self.lowercase:
            result = result.lower()
        if self.uppercase:
            result = result.upper()
        return result


class RangeRule(ValidationRule):
    """Validates numeric ranges."""
    
    def __init__(
        self,
        min: Optional[float] = None,
        max: Optional[float] = None,
        exclusive_min: bool = False,
        exclusive_max: bool = False,
    ):
        self.min = min
        self.max = max
        self.exclusive_min = exclusive_min
        self.exclusive_max = exclusive_max
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None:
            return ValidationResult(valid=True, sanitized_data=value)
        
        if not isinstance(value, (int, float)):
            return ValidationResult(
                valid=False,
                errors=[ValidationError(f"Expected number, got {type(value).__name__}", field_name, value)],
            )
        
        errors = []
        
        if self.min is not None:
            if self.exclusive_min and value <= self.min:
                errors.append(ValidationError(f"Value must be greater than {self.min}", field_name, value))
            elif not self.exclusive_min and value < self.min:
                errors.append(ValidationError(f"Value must be at least {self.min}", field_name, value))
        
        if self.max is not None:
            if self.exclusive_max and value >= self.max:
                errors.append(ValidationError(f"Value must be less than {self.max}", field_name, value))
            elif not self.exclusive_max and value > self.max:
                errors.append(ValidationError(f"Value must be at most {self.max}", field_name, value))
        
        return ValidationResult(
            valid=len(errors) == 0,
            errors=errors,
            sanitized_data=value,
        )


class EmailRule(ValidationRule):
    """Validates email addresses."""
    
    EMAIL_PATTERN = re.compile(
        r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
    )
    
    def __init__(self, normalize: bool = True):
        self.normalize = normalize
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None or value == "":
            return ValidationResult(valid=True, sanitized_data=value)
        
        if not isinstance(value, str):
            return ValidationResult(
                valid=False,
                errors=[ValidationError("Email must be a string", field_name, value)],
            )
        
        sanitized = value.strip().lower() if self.normalize else value
        
        if not self.EMAIL_PATTERN.match(sanitized):
            return ValidationResult(
                valid=False,
                errors=[ValidationError("Invalid email format", field_name, value)],
            )
        
        return ValidationResult(valid=True, sanitized_data=sanitized)


class URLRule(ValidationRule):
    """Validates URLs."""
    
    URL_PATTERN = re.compile(
        r'^https?://[^\s<>"\{\}\|\\^\[\]`]+$'
    )
    
    def __init__(self, allowed_schemes: Optional[List[str]] = None):
        self.allowed_schemes = allowed_schemes or ["http", "https"]
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None or value == "":
            return ValidationResult(valid=True, sanitized_data=value)
        
        if not isinstance(value, str):
            return ValidationResult(
                valid=False,
                errors=[ValidationError("URL must be a string", field_name, value)],
            )
        
        sanitized = value.strip()
        
        # Check scheme
        scheme = sanitized.split("://")[0] if "://" in sanitized else ""
        if scheme not in self.allowed_schemes:
            return ValidationResult(
                valid=False,
                errors=[ValidationError(f"URL scheme must be one of: {self.allowed_schemes}", field_name, value)],
            )
        
        if not self.URL_PATTERN.match(sanitized):
            return ValidationResult(
                valid=False,
                errors=[ValidationError("Invalid URL format", field_name, value)],
            )
        
        return ValidationResult(valid=True, sanitized_data=sanitized)


class EnumRule(ValidationRule):
    """Validates against allowed values."""
    
    def __init__(
        self,
        allowed: List[Any],
        case_insensitive: bool = False,
    ):
        self.allowed = allowed
        self.case_insensitive = case_insensitive
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None:
            return ValidationResult(valid=True, sanitized_data=value)
        
        check_value = value.lower() if self.case_insensitive and isinstance(value, str) else value
        check_allowed = [a.lower() if self.case_insensitive and isinstance(a, str) else a for a in self.allowed]
        
        if check_value not in check_allowed:
            return ValidationResult(
                valid=False,
                errors=[ValidationError(f"Value must be one of: {self.allowed}", field_name, value)],
            )
        
        return ValidationResult(valid=True, sanitized_data=value)


class ListRule(ValidationRule):
    """Validates list values."""
    
    def __init__(
        self,
        item_rule: Optional[ValidationRule] = None,
        min_items: Optional[int] = None,
        max_items: Optional[int] = None,
        unique: bool = False,
    ):
        self.item_rule = item_rule
        self.min_items = min_items
        self.max_items = max_items
        self.unique = unique
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None:
            return ValidationResult(valid=True, sanitized_data=value)
        
        if not isinstance(value, list):
            return ValidationResult(
                valid=False,
                errors=[ValidationError(f"Expected list, got {type(value).__name__}", field_name, value)],
            )
        
        errors = []
        sanitized = []
        
        if self.min_items and len(value) < self.min_items:
            errors.append(ValidationError(f"Minimum {self.min_items} items required", field_name, value))
        
        if self.max_items and len(value) > self.max_items:
            errors.append(ValidationError(f"Maximum {self.max_items} items allowed", field_name, value))
        
        if self.unique:
            try:
                if len(value) != len(set(value)):
                    errors.append(ValidationError("Items must be unique", field_name, value))
            except TypeError:
                pass  # Unhashable items
        
        if self.item_rule:
            for i, item in enumerate(value):
                result = self.item_rule.validate(item, f"{field_name}[{i}]")
                if not result.valid:
                    errors.extend(result.errors)
                sanitized.append(result.sanitized_data if result.sanitized_data is not None else item)
        else:
            sanitized = value
        
        return ValidationResult(
            valid=len(errors) == 0,
            errors=errors,
            sanitized_data=sanitized,
        )


class DictRule(ValidationRule):
    """Validates dictionary values."""
    
    def __init__(
        self,
        schema: Optional[Dict[str, ValidationRule]] = None,
        allow_extra: bool = True,
        remove_extra: bool = False,
    ):
        self.schema = schema or {}
        self.allow_extra = allow_extra
        self.remove_extra = remove_extra
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        if value is None:
            return ValidationResult(valid=True, sanitized_data=value)
        
        if not isinstance(value, dict):
            return ValidationResult(
                valid=False,
                errors=[ValidationError(f"Expected dict, got {type(value).__name__}", field_name, value)],
            )
        
        errors = []
        sanitized = {}
        
        # Validate schema fields
        for key, rule in self.schema.items():
            field_value = value.get(key)
            result = rule.validate(field_value, f"{field_name}.{key}" if field_name else key)
            if not result.valid:
                errors.extend(result.errors)
            sanitized[key] = result.sanitized_data if result.sanitized_data is not None else field_value
        
        # Handle extra fields
        extra_keys = set(value.keys()) - set(self.schema.keys())
        if extra_keys:
            if not self.allow_extra:
                errors.append(ValidationError(f"Unexpected fields: {extra_keys}", field_name, value))
            elif not self.remove_extra:
                for key in extra_keys:
                    sanitized[key] = value[key]
        
        return ValidationResult(
            valid=len(errors) == 0,
            errors=errors,
            sanitized_data=sanitized,
        )


class CustomRule(ValidationRule):
    """Custom validation using a function."""
    
    def __init__(
        self,
        validator: Callable[[Any], bool],
        message: str = "Validation failed",
        sanitizer: Optional[Callable[[Any], Any]] = None,
    ):
        self.validator = validator
        self.message = message
        self.sanitizer = sanitizer
    
    def validate(self, value: Any, field_name: str = "") -> ValidationResult:
        try:
            if not self.validator(value):
                return ValidationResult(
                    valid=False,
                    errors=[ValidationError(self.message, field_name, value)],
                )
        except Exception as e:
            return ValidationResult(
                valid=False,
                errors=[ValidationError(str(e), field_name, value)],
            )
        
        sanitized = self.sanitizer(value) if self.sanitizer else value
        return ValidationResult(valid=True, sanitized_data=sanitized)


class Validator:
    """
    Composite validator for multiple fields.
    """
    
    def __init__(self):
        self._rules: Dict[str, List[ValidationRule]] = {}
        self._required_fields: Set[str] = set()
    
    def add_rule(self, field: str, rule: ValidationRule) -> 'Validator':
        """Add a validation rule for a field."""
        if field not in self._rules:
            self._rules[field] = []
        self._rules[field].append(rule)
        return self
    
    def require(self, *fields: str) -> 'Validator':
        """Mark fields as required."""
        self._required_fields.update(fields)
        return self
    
    def validate(self, data: Dict[str, Any]) -> ValidationResult:
        """
        Validate data against all rules.
        
        Args:
            data: Dictionary to validate
            
        Returns:
            ValidationResult with combined errors
        """
        all_errors = []
        sanitized = {}
        
        # Check required fields
        for field in self._required_fields:
            if field not in data or data[field] is None:
                all_errors.append(ValidationError(f"Field is required", field))
        
        # Validate each field
        for field, rules in self._rules.items():
            value = data.get(field)
            
            for rule in rules:
                result = rule.validate(value, field)
                if not result.valid:
                    all_errors.extend(result.errors)
                else:
                    value = result.sanitized_data
            
            sanitized[field] = value
        
        # Include non-validated fields
        for key, value in data.items():
            if key not in sanitized:
                sanitized[key] = value
        
        return ValidationResult(
            valid=len(all_errors) == 0,
            errors=all_errors,
            sanitized_data=sanitized,
        )


class Schema:
    """
    Declarative schema for validation.
    """
    
    def __init__(self, fields: Dict[str, Any]):
        """
        Initialize schema.
        
        Args:
            fields: Dictionary mapping field names to types or rules
        """
        self._fields = fields
        self._validator = self._build_validator()
    
    def _build_validator(self) -> Validator:
        """Build validator from schema definition."""
        validator = Validator()
        
        for field, spec in self._fields.items():
            if isinstance(spec, ValidationRule):
                validator.add_rule(field, spec)
            elif isinstance(spec, type):
                validator.add_rule(field, TypeRule(spec))
            elif isinstance(spec, dict):
                # Nested schema
                validator.add_rule(field, DictRule(
                    schema={k: TypeRule(v) if isinstance(v, type) else v 
                            for k, v in spec.items()}
                ))
        
        return validator
    
    def validate(self, data: Dict[str, Any]) -> ValidationResult:
        """Validate data against schema."""
        return self._validator.validate(data)


def validate_input(
    schema: Union[Schema, Dict[str, Any], Type],
    raise_on_error: bool = True,
) -> Callable:
    """
    Decorator for input validation.
    
    Example:
        @validate_input({"name": StringRule(min_length=1)})
        async def create_user(data: dict) -> dict:
            return {"id": 1, "name": data["name"]}
    """
    def decorator(func: Callable) -> Callable:
        if isinstance(schema, Schema):
            validator = schema
        elif isinstance(schema, dict):
            validator = Schema(schema)
        else:
            validator = Schema({"data": TypeRule(schema)})
        
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            # Find the data argument
            data = kwargs.get("data") or (args[0] if args else {})
            
            result = validator.validate(data)
            if not result.valid:
                if raise_on_error:
                    raise ValidationError(
                        "Input validation failed",
                        errors=result.errors,
                    )
                return result
            
            # Replace with sanitized data
            if "data" in kwargs:
                kwargs["data"] = result.sanitized_data
            elif args:
                args = (result.sanitized_data,) + args[1:]
            
            return await func(*args, **kwargs)
        
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            data = kwargs.get("data") or (args[0] if args else {})
            
            result = validator.validate(data)
            if not result.valid:
                if raise_on_error:
                    raise ValidationError(
                        "Input validation failed",
                        errors=result.errors,
                    )
                return result
            
            if "data" in kwargs:
                kwargs["data"] = result.sanitized_data
            elif args:
                args = (result.sanitized_data,) + args[1:]
            
            return func(*args, **kwargs)
        
        import asyncio
        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
    
    return decorator


def validate_output(
    schema: Union[Schema, Dict[str, Any]],
    raise_on_error: bool = True,
) -> Callable:
    """
    Decorator for output validation.
    
    Example:
        @validate_output({"id": TypeRule(int), "name": StringRule()})
        async def get_user(user_id: int) -> dict:
            return {"id": user_id, "name": "John"}
    """
    def decorator(func: Callable) -> Callable:
        if isinstance(schema, Schema):
            validator = schema
        else:
            validator = Schema(schema)
        
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            result = await func(*args, **kwargs)
            
            validation = validator.validate(result)
            if not validation.valid:
                if raise_on_error:
                    raise ValidationError(
                        "Output validation failed",
                        errors=validation.errors,
                    )
                logger.error(f"Output validation failed: {validation.errors}")
            
            return validation.sanitized_data or result
        
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            
            validation = validator.validate(result)
            if not validation.valid:
                if raise_on_error:
                    raise ValidationError(
                        "Output validation failed",
                        errors=validation.errors,
                    )
                logger.error(f"Output validation failed: {validation.errors}")
            
            return validation.sanitized_data or result
        
        import asyncio
        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
    
    return decorator


__all__ = [
    # Exceptions
    "ValidationError",
    # Data classes
    "ValidationResult",
    # Rules
    "ValidationRule",
    "RequiredRule",
    "TypeRule",
    "StringRule",
    "RangeRule",
    "EmailRule",
    "URLRule",
    "EnumRule",
    "ListRule",
    "DictRule",
    "CustomRule",
    # Validators
    "Validator",
    "Schema",
    # Decorators
    "validate_input",
    "validate_output",
]



"""
Enterprise API Client Module.

Provides HTTP client abstraction with retry, circuit breaker,
request/response interceptors, and resilience patterns.

Example:
    # Create API client
    client = create_api_client(
        base_url="https://api.example.com",
        timeout=30.0,
    )
    
    # Make requests
    response = await client.get("/users/123")
    response = await client.post("/users", json={"name": "John"})
    
    # Use decorators
    @api_client("user-service")
    async def fetch_user(client: ApiClient, user_id: str):
        return await client.get(f"/users/{user_id}")
"""

from __future__ import annotations

import asyncio
import functools
import hashlib
import json
import logging
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class ApiClientError(Exception):
    """Base API client error."""
    pass


class ConnectionError(ApiClientError):
    """Connection failed."""
    pass


class TimeoutError(ApiClientError):
    """Request timed out."""
    pass


class HttpError(ApiClientError):
    """HTTP error response."""
    def __init__(
        self,
        message: str,
        status_code: int,
        response: Optional["Response"] = None,
    ):
        super().__init__(message)
        self.status_code = status_code
        self.response = response


class RetryExhaustedError(ApiClientError):
    """All retries exhausted."""
    pass


class HttpMethod(str, Enum):
    """HTTP methods."""
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"


class RetryStrategy(str, Enum):
    """Retry strategy."""
    NONE = "none"
    FIXED = "fixed"
    EXPONENTIAL = "exponential"
    EXPONENTIAL_JITTER = "exponential_jitter"


@dataclass
class RetryConfig:
    """Retry configuration."""
    strategy: RetryStrategy = RetryStrategy.EXPONENTIAL_JITTER
    max_retries: int = 3
    base_delay: float = 1.0
    max_delay: float = 30.0
    retry_on: Set[int] = field(default_factory=lambda: {500, 502, 503, 504})
    retry_on_timeout: bool = True
    retry_on_connection_error: bool = True


@dataclass
class CircuitBreakerConfig:
    """Circuit breaker configuration."""
    enabled: bool = True
    failure_threshold: int = 5
    recovery_timeout: float = 30.0
    half_open_requests: int = 1


@dataclass
class ClientConfig:
    """API client configuration."""
    base_url: str = ""
    timeout: float = 30.0
    connect_timeout: float = 10.0
    headers: Dict[str, str] = field(default_factory=dict)
    retry: RetryConfig = field(default_factory=RetryConfig)
    circuit_breaker: CircuitBreakerConfig = field(default_factory=CircuitBreakerConfig)
    verify_ssl: bool = True
    follow_redirects: bool = True
    max_redirects: int = 10


@dataclass
class Request:
    """HTTP request."""
    method: HttpMethod
    url: str
    headers: Dict[str, str] = field(default_factory=dict)
    params: Dict[str, Any] = field(default_factory=dict)
    json: Optional[Dict[str, Any]] = None
    data: Optional[Any] = None
    files: Optional[Dict[str, Any]] = None
    timeout: Optional[float] = None
    
    # Metadata
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    timestamp: datetime = field(default_factory=datetime.utcnow)
    
    def full_url(self, base_url: str = "") -> str:
        """Get full URL with base."""
        if self.url.startswith(("http://", "https://")):
            return self.url
        return f"{base_url.rstrip('/')}/{self.url.lstrip('/')}"


@dataclass
class Response:
    """HTTP response."""
    status_code: int
    headers: Dict[str, str] = field(default_factory=dict)
    body: bytes = b""
    json_data: Optional[Dict[str, Any]] = None
    request: Optional[Request] = None
    elapsed: float = 0.0
    
    @property
    def ok(self) -> bool:
        """Check if response is successful."""
        return 200 <= self.status_code < 300
    
    @property
    def text(self) -> str:
        """Get response as text."""
        return self.body.decode("utf-8", errors="replace")
    
    def json(self) -> Any:
        """Parse response as JSON."""
        if self.json_data is not None:
            return self.json_data
        return json.loads(self.body)
    
    def raise_for_status(self) -> None:
        """Raise exception for error status codes."""
        if not self.ok:
            raise HttpError(
                f"HTTP {self.status_code}",
                status_code=self.status_code,
                response=self,
            )


class RequestInterceptor(ABC):
    """Request interceptor."""
    
    @abstractmethod
    async def intercept(self, request: Request) -> Request:
        """Intercept and modify request."""
        pass


class ResponseInterceptor(ABC):
    """Response interceptor."""
    
    @abstractmethod
    async def intercept(
        self,
        response: Response,
        request: Request,
    ) -> Response:
        """Intercept and modify response."""
        pass


class AuthInterceptor(RequestInterceptor):
    """Authentication interceptor."""
    
    def __init__(
        self,
        token_provider: Callable[[], Awaitable[str]],
        header_name: str = "Authorization",
        prefix: str = "Bearer",
    ):
        self._token_provider = token_provider
        self._header_name = header_name
        self._prefix = prefix
    
    async def intercept(self, request: Request) -> Request:
        token = await self._token_provider()
        request.headers[self._header_name] = f"{self._prefix} {token}"
        return request


class LoggingInterceptor(RequestInterceptor, ResponseInterceptor):
    """Logging interceptor."""
    
    def __init__(self, log_body: bool = False):
        self._log_body = log_body
    
    async def intercept(self, request: Request) -> Request:
        logger.info(f"Request: {request.method} {request.url}")
        if self._log_body and request.json:
            logger.debug(f"Request body: {request.json}")
        return request
    
    async def intercept(
        self,
        response: Response,
        request: Request,
    ) -> Response:
        logger.info(
            f"Response: {response.status_code} "
            f"({response.elapsed:.3f}s)"
        )
        return response


class RetryInterceptor(ResponseInterceptor):
    """Retry response interceptor for specific conditions."""
    
    def __init__(
        self,
        should_retry: Callable[[Response], bool],
        max_retries: int = 3,
    ):
        self._should_retry = should_retry
        self._max_retries = max_retries
    
    async def intercept(
        self,
        response: Response,
        request: Request,
    ) -> Response:
        # Just mark for retry, actual retry handled by client
        return response


class CircuitBreaker:
    """Circuit breaker for API calls."""
    
    def __init__(self, config: CircuitBreakerConfig):
        self._config = config
        self._failures = 0
        self._last_failure: Optional[datetime] = None
        self._state = "closed"  # closed, open, half-open
    
    @property
    def state(self) -> str:
        return self._state
    
    def is_available(self) -> bool:
        """Check if circuit allows requests."""
        if self._state == "closed":
            return True
        
        if self._state == "open":
            # Check if recovery timeout has passed
            if self._last_failure:
                elapsed = (datetime.utcnow() - self._last_failure).total_seconds()
                if elapsed >= self._config.recovery_timeout:
                    self._state = "half-open"
                    return True
            return False
        
        # Half-open: allow limited requests
        return True
    
    def record_success(self) -> None:
        """Record successful call."""
        if self._state == "half-open":
            self._state = "closed"
            self._failures = 0
    
    def record_failure(self) -> None:
        """Record failed call."""
        self._failures += 1
        self._last_failure = datetime.utcnow()
        
        if self._failures >= self._config.failure_threshold:
            self._state = "open"
            logger.warning("Circuit breaker opened")
    
    def reset(self) -> None:
        """Reset circuit breaker."""
        self._state = "closed"
        self._failures = 0
        self._last_failure = None


class HttpTransport(ABC):
    """Abstract HTTP transport."""
    
    @abstractmethod
    async def send(self, request: Request) -> Response:
        """Send HTTP request."""
        pass


class MockHttpTransport(HttpTransport):
    """Mock HTTP transport for testing."""
    
    def __init__(self):
        self._responses: Dict[str, Response] = {}
        self._default_response = Response(status_code=200)
    
    def mock(
        self,
        method: HttpMethod,
        url: str,
        response: Response,
    ) -> None:
        """Mock a response."""
        key = f"{method}:{url}"
        self._responses[key] = response
    
    async def send(self, request: Request) -> Response:
        key = f"{request.method}:{request.url}"
        response = self._responses.get(key, self._default_response)
        response.request = request
        return response


class ApiClient:
    """
    HTTP API client with resilience patterns.
    """
    
    def __init__(
        self,
        config: Optional[ClientConfig] = None,
        transport: Optional[HttpTransport] = None,
    ):
        self._config = config or ClientConfig()
        self._transport = transport or MockHttpTransport()
        
        self._request_interceptors: List[RequestInterceptor] = []
        self._response_interceptors: List[ResponseInterceptor] = []
        
        self._circuit_breaker = CircuitBreaker(self._config.circuit_breaker)
    
    @property
    def base_url(self) -> str:
        return self._config.base_url
    
    def add_request_interceptor(self, interceptor: RequestInterceptor) -> None:
        """Add request interceptor."""
        self._request_interceptors.append(interceptor)
    
    def add_response_interceptor(self, interceptor: ResponseInterceptor) -> None:
        """Add response interceptor."""
        self._response_interceptors.append(interceptor)
    
    async def request(
        self,
        method: HttpMethod,
        url: str,
        **kwargs,
    ) -> Response:
        """Make HTTP request."""
        request = Request(
            method=method,
            url=url,
            headers={**self._config.headers, **kwargs.pop("headers", {})},
            params=kwargs.pop("params", {}),
            json=kwargs.pop("json", None),
            data=kwargs.pop("data", None),
            timeout=kwargs.pop("timeout", self._config.timeout),
        )
        
        # Apply request interceptors
        for interceptor in self._request_interceptors:
            request = await interceptor.intercept(request)
        
        # Check circuit breaker
        if self._config.circuit_breaker.enabled:
            if not self._circuit_breaker.is_available():
                raise ApiClientError("Circuit breaker is open")
        
        # Retry loop
        retry_config = self._config.retry
        attempt = 0
        last_exception: Optional[Exception] = None
        
        while attempt <= retry_config.max_retries:
            try:
                start_time = time.time()
                response = await self._send_request(request)
                response.elapsed = time.time() - start_time
                
                # Apply response interceptors
                for interceptor in self._response_interceptors:
                    response = await interceptor.intercept(response, request)
                
                # Check if should retry
                if (
                    not response.ok and
                    response.status_code in retry_config.retry_on and
                    attempt < retry_config.max_retries
                ):
                    attempt += 1
                    await self._wait_for_retry(attempt, retry_config)
                    continue
                
                # Record success
                if self._config.circuit_breaker.enabled:
                    self._circuit_breaker.record_success()
                
                return response
                
            except (ConnectionError, TimeoutError) as e:
                last_exception = e
                
                # Record failure
                if self._config.circuit_breaker.enabled:
                    self._circuit_breaker.record_failure()
                
                if attempt < retry_config.max_retries:
                    if (
                        (isinstance(e, TimeoutError) and retry_config.retry_on_timeout) or
                        (isinstance(e, ConnectionError) and retry_config.retry_on_connection_error)
                    ):
                        attempt += 1
                        await self._wait_for_retry(attempt, retry_config)
                        continue
                
                raise
        
        if last_exception:
            raise RetryExhaustedError(f"Retries exhausted: {last_exception}")
        
        raise RetryExhaustedError("All retries exhausted")
    
    async def _send_request(self, request: Request) -> Response:
        """Send request via transport."""
        return await self._transport.send(request)
    
    async def _wait_for_retry(
        self,
        attempt: int,
        config: RetryConfig,
    ) -> None:
        """Wait before retry."""
        if config.strategy == RetryStrategy.NONE:
            return
        
        if config.strategy == RetryStrategy.FIXED:
            delay = config.base_delay
        elif config.strategy == RetryStrategy.EXPONENTIAL:
            delay = min(
                config.base_delay * (2 ** (attempt - 1)),
                config.max_delay
            )
        else:  # EXPONENTIAL_JITTER
            import random
            base = min(
                config.base_delay * (2 ** (attempt - 1)),
                config.max_delay
            )
            delay = base * (0.5 + random.random())
        
        logger.debug(f"Retry attempt {attempt}, waiting {delay:.2f}s")
        await asyncio.sleep(delay)
    
    # Convenience methods
    async def get(
        self,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Response:
        """Make GET request."""
        return await self.request(
            HttpMethod.GET,
            url,
            params=params or {},
            **kwargs,
        )
    
    async def post(
        self,
        url: str,
        json: Optional[Dict[str, Any]] = None,
        data: Optional[Any] = None,
        **kwargs,
    ) -> Response:
        """Make POST request."""
        return await self.request(
            HttpMethod.POST,
            url,
            json=json,
            data=data,
            **kwargs,
        )
    
    async def put(
        self,
        url: str,
        json: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Response:
        """Make PUT request."""
        return await self.request(
            HttpMethod.PUT,
            url,
            json=json,
            **kwargs,
        )
    
    async def patch(
        self,
        url: str,
        json: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Response:
        """Make PATCH request."""
        return await self.request(
            HttpMethod.PATCH,
            url,
            json=json,
            **kwargs,
        )
    
    async def delete(
        self,
        url: str,
        **kwargs,
    ) -> Response:
        """Make DELETE request."""
        return await self.request(HttpMethod.DELETE, url, **kwargs)


class ServiceApiClient(ApiClient):
    """
    API client for service-to-service communication.
    """
    
    def __init__(
        self,
        service_name: str,
        config: Optional[ClientConfig] = None,
        service_registry: Optional[Any] = None,
    ):
        super().__init__(config)
        self._service_name = service_name
        self._service_registry = service_registry
    
    async def discover_base_url(self) -> str:
        """Discover service base URL."""
        if self._service_registry:
            instance = await self._service_registry.discover(self._service_name)
            return instance.base_url
        return self._config.base_url


# Global clients
_clients: Dict[str, ApiClient] = {}


# Decorators
def api_client(
    service_name: str,
    base_url: Optional[str] = None,
) -> Callable:
    """
    Decorator to inject API client.
    
    Example:
        @api_client("user-service", base_url="https://api.example.com")
        async def get_user(client: ApiClient, user_id: str):
            return await client.get(f"/users/{user_id}")
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            client = get_client(service_name, base_url)
            return await func(client, *args, **kwargs)
        
        return wrapper
    
    return decorator


def with_retry(
    max_retries: int = 3,
    retry_on: Optional[Set[int]] = None,
) -> Callable:
    """
    Decorator to add retry logic.
    
    Example:
        @with_retry(max_retries=5)
        async def call_api():
            ...
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            last_error = None
            
            for attempt in range(max_retries + 1):
                try:
                    return await func(*args, **kwargs)
                except HttpError as e:
                    if retry_on and e.status_code not in retry_on:
                        raise
                    last_error = e
                    
                    if attempt < max_retries:
                        await asyncio.sleep(2 ** attempt)
                except Exception as e:
                    last_error = e
                    
                    if attempt < max_retries:
                        await asyncio.sleep(2 ** attempt)
            
            raise RetryExhaustedError(f"Retries exhausted: {last_error}")
        
        return wrapper
    
    return decorator


# Factory functions
def create_api_client(
    base_url: str = "",
    timeout: float = 30.0,
    headers: Optional[Dict[str, str]] = None,
    max_retries: int = 3,
) -> ApiClient:
    """Create API client."""
    config = ClientConfig(
        base_url=base_url,
        timeout=timeout,
        headers=headers or {},
        retry=RetryConfig(max_retries=max_retries),
    )
    return ApiClient(config)


def create_service_client(
    service_name: str,
    base_url: str = "",
    service_registry: Optional[Any] = None,
) -> ServiceApiClient:
    """Create service API client."""
    config = ClientConfig(base_url=base_url)
    return ServiceApiClient(service_name, config, service_registry)


def create_client_config(
    base_url: str = "",
    timeout: float = 30.0,
    max_retries: int = 3,
) -> ClientConfig:
    """Create client configuration."""
    return ClientConfig(
        base_url=base_url,
        timeout=timeout,
        retry=RetryConfig(max_retries=max_retries),
    )


def create_auth_interceptor(
    token_provider: Callable[[], Awaitable[str]],
    header_name: str = "Authorization",
    prefix: str = "Bearer",
) -> AuthInterceptor:
    """Create auth interceptor."""
    return AuthInterceptor(token_provider, header_name, prefix)


def create_logging_interceptor(log_body: bool = False) -> LoggingInterceptor:
    """Create logging interceptor."""
    return LoggingInterceptor(log_body)


def create_mock_transport() -> MockHttpTransport:
    """Create mock transport for testing."""
    return MockHttpTransport()


def get_client(
    name: str,
    base_url: Optional[str] = None,
) -> ApiClient:
    """Get or create named client."""
    if name not in _clients:
        _clients[name] = create_api_client(base_url or "")
    return _clients[name]


__all__ = [
    # Exceptions
    "ApiClientError",
    "ConnectionError",
    "TimeoutError",
    "HttpError",
    "RetryExhaustedError",
    # Enums
    "HttpMethod",
    "RetryStrategy",
    # Data classes
    "RetryConfig",
    "CircuitBreakerConfig",
    "ClientConfig",
    "Request",
    "Response",
    # Interceptors
    "RequestInterceptor",
    "ResponseInterceptor",
    "AuthInterceptor",
    "LoggingInterceptor",
    "RetryInterceptor",
    # Circuit breaker
    "CircuitBreaker",
    # Transport
    "HttpTransport",
    "MockHttpTransport",
    # Clients
    "ApiClient",
    "ServiceApiClient",
    # Decorators
    "api_client",
    "with_retry",
    # Factory functions
    "create_api_client",
    "create_service_client",
    "create_client_config",
    "create_auth_interceptor",
    "create_logging_interceptor",
    "create_mock_transport",
    "get_client",
]



"""
Memory and RAG Evaluation systems.

Provides:
- MemoryEvaluator: Memory accuracy, context relevance, knowledge freshness
- RAGEvaluator: Retrieval accuracy, citation correctness, answer groundedness
"""

import uuid
import time
import logging
import statistics
from typing import Dict, Any, List
from collections import defaultdict

logger = logging.getLogger(__name__)


class MemoryEvaluator:
    """Memory & Context Evaluation system."""
    
    def __init__(self):
        self.memory_evaluations: List[Dict[str, Any]] = []
        self.memory_metrics: Dict[str, Any] = {
            'total_queries': 0,
            'relevant_retrievals': 0,
            'stale_data_usage': 0,
            'context_precision_scores': [],
            'memory_overwrite_errors': 0
        }
    
    def evaluate_memory_retrieval(self,
                                  query: str,
                                  retrieved_memories: List[Dict[str, Any]],
                                  relevant_memories: List[Dict[str, Any]] = None,
                                  _metadata: Dict[str, Any] = None) -> Dict[str, Any]:
        """Evaluate memory retrieval quality."""
        evaluation = {
            'id': str(uuid.uuid4()),
            'query': query,
            'retrieved_count': len(retrieved_memories),
            'timestamp': time.time()
        }
        
        if relevant_memories:
            precision = self._calculate_precision(retrieved_memories, relevant_memories)
            recall = self._calculate_recall(retrieved_memories, relevant_memories)
            evaluation['precision'] = precision
            evaluation['recall'] = recall
            self.memory_metrics['context_precision_scores'].append(precision)
        
        stale_count = sum(1 for m in retrieved_memories if self._is_stale(m))
        evaluation['stale_count'] = stale_count
        
        if stale_count > 0:
            self.memory_metrics['stale_data_usage'] += 1
        
        self.memory_evaluations.append(evaluation)
        self.memory_metrics['total_queries'] += 1
        
        return evaluation
    
    def _calculate_precision(self, retrieved: List[Dict], relevant: List[Dict]) -> float:
        if not retrieved:
            return 0.0
        retrieved_ids = {m.get('id', str(m)) for m in retrieved}
        relevant_ids = {m.get('id', str(m)) for m in relevant}
        return len(retrieved_ids & relevant_ids) / len(retrieved_ids)
    
    def _calculate_recall(self, retrieved: List[Dict], relevant: List[Dict]) -> float:
        if not relevant:
            return 0.0
        retrieved_ids = {m.get('id', str(m)) for m in retrieved}
        relevant_ids = {m.get('id', str(m)) for m in relevant}
        return len(retrieved_ids & relevant_ids) / len(relevant_ids)
    
    def _is_stale(self, memory: Dict[str, Any]) -> bool:
        if 'timestamp' not in memory:
            return False
        age_seconds = time.time() - memory['timestamp']
        return age_seconds > (30 * 24 * 3600)
    
    def record_memory_error(self, error_type: str):
        """Record memory-related errors."""
        if error_type == 'overwrite':
            self.memory_metrics['memory_overwrite_errors'] += 1
    
    def get_memory_metrics(self) -> Dict[str, Any]:
        """Get memory evaluation metrics."""
        metrics = self.memory_metrics
        return {
            'total_queries': metrics['total_queries'],
            'stale_data_rate': metrics['stale_data_usage'] / metrics['total_queries'] if metrics['total_queries'] else 0,
            'avg_precision': statistics.mean(metrics['context_precision_scores']) if metrics['context_precision_scores'] else 0,
            'memory_errors': metrics['memory_overwrite_errors']
        }
    
    def evaluate_retrieval(self, query: str, retrieved_memories: List[Dict[str, Any]],
                          relevant_memories: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Evaluate memory retrieval (alias)."""
        return self.evaluate_memory_retrieval(query, retrieved_memories, relevant_memories)
    
    def record_stale_data_access(self, memory_data: Dict[str, Any] = None, _memory_data: Dict[str, Any] = None):
        """Record access to stale data."""
        # Support both memory_data and _memory_data for compatibility
        self.memory_metrics['stale_data_usage'] += 1
    
    def record_overwrite_error(self):
        """Record a memory overwrite error."""
        self.record_memory_error('overwrite')


class RAGEvaluator:
    """RAG (Retrieval-Augmented Generation) Evaluation system."""
    
    def __init__(self):
        self.rag_evaluations: List[Dict[str, Any]] = []
        self.rag_metrics: Dict[str, Any] = {
            'total_queries': 0,
            'precision_scores': [],
            'recall_scores': [],
            'faithfulness_scores': [],
            'groundedness_scores': []
        }
    
    def evaluate_rag_response(self,
                             query: str,
                             retrieved_docs: List[Dict[str, Any]],
                             generated_answer: str,
                             relevant_docs: List[Dict[str, Any]] = None,
                             _ground_truth_answer: str = None,
                             _metadata: Dict[str, Any] = None) -> Dict[str, Any]:
        """Evaluate RAG response quality."""
        evaluation = {
            'id': str(uuid.uuid4()),
            'query': query,
            'num_retrieved': len(retrieved_docs),
            'answer_length': len(generated_answer),
            'timestamp': time.time()
        }
        
        if relevant_docs:
            precision = self._calculate_retrieval_precision(retrieved_docs, relevant_docs)
            recall = self._calculate_retrieval_recall(retrieved_docs, relevant_docs)
            evaluation['retrieval_precision'] = precision
            evaluation['retrieval_recall'] = recall
            self.rag_metrics['precision_scores'].append(precision)
            self.rag_metrics['recall_scores'].append(recall)
        
        faithfulness = self._assess_faithfulness(generated_answer, retrieved_docs)
        evaluation['faithfulness'] = faithfulness
        self.rag_metrics['faithfulness_scores'].append(faithfulness)
        
        groundedness = self._assess_groundedness(generated_answer, retrieved_docs)
        evaluation['groundedness'] = groundedness
        self.rag_metrics['groundedness_scores'].append(groundedness)
        
        evaluation['has_citations'] = self._has_citations(generated_answer)
        
        self.rag_evaluations.append(evaluation)
        self.rag_metrics['total_queries'] += 1
        
        return evaluation
    
    def _calculate_retrieval_precision(self, retrieved: List[Dict], relevant: List[Dict]) -> float:
        if not retrieved:
            return 0.0
        retrieved_ids = {str(d.get('id', d)) for d in retrieved}
        relevant_ids = {str(d.get('id', d)) for d in relevant}
        return len(retrieved_ids & relevant_ids) / len(retrieved_ids)
    
    def _calculate_retrieval_recall(self, retrieved: List[Dict], relevant: List[Dict]) -> float:
        if not relevant:
            return 0.0
        retrieved_ids = {str(d.get('id', d)) for d in retrieved}
        relevant_ids = {str(d.get('id', d)) for d in relevant}
        return len(retrieved_ids & relevant_ids) / len(relevant_ids)
    
    def _assess_faithfulness(self, answer: str, docs: List[Dict]) -> float:
        if not docs:
            return 0.0
        answer_tokens = set(answer.lower().split())
        doc_tokens = set()
        for doc in docs:
            content = doc.get('content', str(doc))
            doc_tokens.update(content.lower().split())
        if not doc_tokens:
            return 0.0
        overlap = len(answer_tokens & doc_tokens)
        return overlap / len(answer_tokens) if answer_tokens else 0.0
    
    def _assess_groundedness(self, answer: str, docs: List[Dict]) -> float:
        return self._assess_faithfulness(answer, docs)
    
    def _has_citations(self, answer: str) -> bool:
        citation_patterns = ['[', ']', 'source:', 'reference:', 'according to']
        return any(pattern in answer.lower() for pattern in citation_patterns)
    
    def get_rag_metrics(self) -> Dict[str, Any]:
        """Get RAG evaluation metrics."""
        metrics = self.rag_metrics
        return {
            'total_queries': metrics['total_queries'],
            'avg_retrieval_precision': statistics.mean(metrics['precision_scores']) if metrics['precision_scores'] else 0,
            'avg_retrieval_recall': statistics.mean(metrics['recall_scores']) if metrics['recall_scores'] else 0,
            'avg_faithfulness': statistics.mean(metrics['faithfulness_scores']) if metrics['faithfulness_scores'] else 0,
            'avg_groundedness': statistics.mean(metrics['groundedness_scores']) if metrics['groundedness_scores'] else 0
        }
    
    def evaluate_retrieval(self, query: str, retrieved_docs: List[Dict[str, Any]],
                          relevant_docs: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Evaluate retrieval quality."""
        precision = self._calculate_retrieval_precision(retrieved_docs, relevant_docs)
        recall = self._calculate_retrieval_recall(retrieved_docs, relevant_docs)
        return {
            'query': query,
            'precision': precision,
            'recall': recall,
            'f1_score': 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0,
            'num_retrieved': len(retrieved_docs),
            'num_relevant': len(relevant_docs),
            'timestamp': time.time()
        }
    
    def evaluate_faithfulness(self, answer: str, retrieved_docs: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Evaluate answer faithfulness."""
        return {
            'answer': answer,
            'faithfulness_score': self._assess_faithfulness(answer, retrieved_docs),
            'num_docs': len(retrieved_docs),
            'timestamp': time.time()
        }
    
    def evaluate_groundedness(self, answer: str, retrieved_docs: List[Dict[str, Any]]) -> Dict[str, Any]:
        """Evaluate answer groundedness."""
        return {
            'answer': answer,
            'groundedness_score': self._assess_groundedness(answer, retrieved_docs),
            'num_docs': len(retrieved_docs),
            'timestamp': time.time()
        }
    
    def check_citations(self, answer: str) -> Dict[str, Any]:
        """Check if answer contains citations."""
        return {
            'answer': answer,
            'has_citations': self._has_citations(answer),
            'timestamp': time.time()
        }


__all__ = ['MemoryEvaluator', 'RAGEvaluator']
